# Copyright (C) 2013-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import logging
import numbers

import numpy

from ffcx.codegeneration.C.format_lines import Indented, format_indented_lines
from ffcx.codegeneration.C.format_value import (format_float, format_int,
                                                format_value)
from ffcx.codegeneration.C.precedence import PRECEDENCE

logger = logging.getLogger("ffcx")
"""CNode TODO:
- Array copy statement
- Extend ArrayDecl and ArrayAccess with support for
  flattened but conceptually multidimensional arrays,
  maybe even with padding (FlattenedArray possibly covers what we need)
- Function declaration
- TypeDef
- Type
- TemplateArgumentList
- Class declaration
- Class definition
"""

# Some helper functions


def is_zero_cexpr(cexpr):
    return ((isinstance(cexpr, LiteralFloat) and cexpr.value == 0.0)
            or (isinstance(cexpr, LiteralInt) and cexpr.value == 0))


def is_one_cexpr(cexpr):
    return ((isinstance(cexpr, LiteralFloat) and cexpr.value == 1.0)
            or (isinstance(cexpr, LiteralInt) and cexpr.value == 1))


def is_negative_one_cexpr(cexpr):
    return ((isinstance(cexpr, LiteralFloat) and cexpr.value == -1.0)
            or (isinstance(cexpr, LiteralInt) and cexpr.value == -1))


def float_product(factors):
    """Build product of float factors, simplifying ones and zeros and returning 1.0 if empty sequence."""
    factors = [f for f in factors if not is_one_cexpr(f)]
    if len(factors) == 0:
        return LiteralFloat(1.0)
    elif len(factors) == 1:
        return factors[0]
    else:
        for f in factors:
            if is_zero_cexpr(f):
                return f
        return Product(factors)
# CNode core


class CNode(object):
    """Base class for all C AST nodes."""

    __slots__ = ()

    def __str__(self):
        name = self.__class__.__name__
        raise NotImplementedError("Missing implementation of __str__ in " + name)

    def __eq__(self, other):
        name = self.__class__.__name__
        raise NotImplementedError("Missing implementation of __eq__ in " + name)

    def __ne__(self, other):
        return not self.__eq__(other)

# CExpr base classes


class CExpr(CNode):
    """Base class for all C expressions.

    All subtypes should define a 'precedence' class attribute.

    """

    __slots__ = ()

    def ce_format(self, precision=None):
        raise NotImplementedError("Missing implementation of ce_format() in CExpr.")

    def __str__(self):
        try:
            s = self.ce_format()
        except Exception:
            raise

        return s

    def __getitem__(self, indices):
        return ArrayAccess(self, indices)

    def __neg__(self):
        if isinstance(self, LiteralFloat):
            return LiteralFloat(-self.value)
        if isinstance(self, LiteralInt):
            return LiteralInt(-self.value)
        return Neg(self)

    def __add__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return other
        if is_zero_cexpr(other):
            return self
        if isinstance(other, Neg):
            return Sub(self, other.arg)
        return Add(self, other)

    def __radd__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return other
        if is_zero_cexpr(other):
            return self
        if isinstance(self, Neg):
            return Sub(other, self.arg)
        return Add(other, self)

    def __sub__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return -other
        if is_zero_cexpr(other):
            return self
        if isinstance(other, Neg):
            return Add(self, other.arg)
        return Sub(self, other)

    def __rsub__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return other
        if is_zero_cexpr(other):
            return -self
        if isinstance(self, Neg):
            return Add(other, self.arg)
        return Sub(other, self)

    def __mul__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return self
        if is_zero_cexpr(other):
            return other
        if is_one_cexpr(self):
            return other
        if is_one_cexpr(other):
            return self
        if is_negative_one_cexpr(other):
            return Neg(self)
        if is_negative_one_cexpr(self):
            return Neg(other)
        return Mul(self, other)

    def __rmul__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            return self
        if is_zero_cexpr(other):
            return other
        if is_one_cexpr(self):
            return other
        if is_one_cexpr(other):
            return self
        if is_negative_one_cexpr(other):
            return Neg(self)
        if is_negative_one_cexpr(self):
            return Neg(other)
        return Mul(other, self)

    def __div__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(other):
            raise ValueError("Division by zero!")
        if is_zero_cexpr(self):
            return self
        return Div(self, other)

    def __rdiv__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            raise ValueError("Division by zero!")
        if is_zero_cexpr(other):
            return other
        return Div(other, self)

    # TODO: Error check types? Can't do that exactly as symbols here have no type.
    __truediv__ = __div__
    __rtruediv__ = __rdiv__
    __floordiv__ = __div__
    __rfloordiv__ = __rdiv__

    def __mod__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(other):
            raise ValueError("Division by zero!")
        if is_zero_cexpr(self):
            return self
        return Mod(self, other)

    def __rmod__(self, other):
        other = as_cexpr(other)
        if is_zero_cexpr(self):
            raise ValueError("Division by zero!")
        if is_zero_cexpr(other):
            return other
        return Mod(other, self)


class CExprOperator(CExpr):
    """Base class for all C expression operator."""

    __slots__ = ()
    sideeffect = False


class CExprTerminal(CExpr):
    """Base class for all C expression terminals."""

    __slots__ = ()
    sideeffect = False


# CExprTerminal types


class CExprLiteral(CExprTerminal):
    """A float or int literal value."""

    __slots__ = ()
    precedence = PRECEDENCE.LITERAL


class Null(CExprLiteral):
    """A null pointer literal."""

    __slots__ = ()
    precedence = PRECEDENCE.LITERAL

    def ce_format(self, precision=None):
        return "NULL"

    def __eq__(self, other):
        return isinstance(other, Null)


class LiteralFloat(CExprLiteral):
    """A floating point literal value."""

    __slots__ = ("value", )
    precedence = PRECEDENCE.LITERAL

    def __init__(self, value):
        assert isinstance(value, (float, complex, int, numpy.number))
        self.value = value

    def ce_format(self, precision=None):
        return format_float(self.value, precision)

    def __eq__(self, other):
        return isinstance(other, LiteralFloat) and self.value == other.value

    def __bool__(self):
        return bool(self.value)

    __nonzero__ = __bool__

    def __float__(self):
        return float(self.value)

    def flops(self):
        return 0


class LiteralInt(CExprLiteral):
    """An integer literal value."""

    __slots__ = ("value", )
    precedence = PRECEDENCE.LITERAL

    def __init__(self, value):
        assert isinstance(value, (int, numpy.number))
        self.value = value

    def ce_format(self, precision=None):
        return str(self.value)

    def flops(self):
        return 0

    def __eq__(self, other):
        return isinstance(other, LiteralInt) and self.value == other.value

    def __bool__(self):
        return bool(self.value)

    __nonzero__ = __bool__

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __hash__(self):
        return hash(self.ce_format())


class LiteralBool(CExprLiteral):
    """A boolean literal value."""

    __slots__ = ("value", )
    precedence = PRECEDENCE.LITERAL

    def __init__(self, value):
        assert isinstance(value, (bool, ))
        self.value = value

    def ce_format(self, precision=None):
        return "true" if self.value else "false"

    def __eq__(self, other):
        return isinstance(other, LiteralBool) and self.value == other.value

    def __bool__(self):
        return bool(self.value)

    __nonzero__ = __bool__


class LiteralString(CExprLiteral):
    """A boolean literal value."""

    __slots__ = ("value", )
    precedence = PRECEDENCE.LITERAL

    def __init__(self, value):
        assert isinstance(value, (str, ))
        assert '"' not in value
        self.value = value

    def ce_format(self, precision=None):
        return '"%s"' % (self.value, )

    def __eq__(self, other):
        return isinstance(other, LiteralString) and self.value == other.value


class Symbol(CExprTerminal):
    """A named symbol."""

    __slots__ = ("name", )
    precedence = PRECEDENCE.SYMBOL

    def __init__(self, name):
        assert isinstance(name, str)
        self.name = name

    def ce_format(self, precision=None):
        return self.name

    def flops(self):
        return 0

    def __eq__(self, other):
        return isinstance(other, Symbol) and self.name == other.name

    def __hash__(self):
        return hash(self.ce_format())


# CExprOperator base classes

class UnaryOp(CExprOperator):
    """Base class for unary operators."""

    __slots__ = ("arg", )

    def __init__(self, arg):
        self.arg = as_cexpr(arg)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.arg == other.arg

    def flops(self):
        raise NotImplementedError()


class PrefixUnaryOp(UnaryOp):
    """Base class for prefix unary operators."""

    __slots__ = ()

    def ce_format(self, precision=None):
        arg = self.arg.ce_format(precision)
        if self.arg.precedence >= self.precedence:
            arg = '(' + arg + ')'
        return self.op + arg

    def __eq__(self, other):
        return isinstance(other, type(self))


class PostfixUnaryOp(UnaryOp):
    """Base class for postfix unary operators."""

    __slots__ = ()

    def ce_format(self, precision=None):
        arg = self.arg.ce_format(precision)
        if self.arg.precedence >= self.precedence:
            arg = '(' + arg + ')'
        return arg + self.op

    def __eq__(self, other):
        return isinstance(other, type(self))


class BinOp(CExprOperator):
    __slots__ = ("lhs", "rhs")

    def __init__(self, lhs, rhs):
        self.lhs = as_cexpr(lhs)
        self.rhs = as_cexpr(rhs)

    def ce_format(self, precision=None):
        # Format children
        lhs = self.lhs.ce_format(precision)
        rhs = self.rhs.ce_format(precision)

        # Apply parentheses
        if self.lhs.precedence >= self.precedence:
            lhs = '(' + lhs + ')'
        if self.rhs.precedence >= self.precedence:
            rhs = '(' + rhs + ')'

        # Return combined string
        return lhs + (" " + self.op + " ") + rhs

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.lhs == other.lhs and self.rhs == other.rhs)

    def __hash__(self):
        return hash(self.ce_format())

    def flops(self):
        return 1 + self.lhs.flops() + self.rhs.flops()


class NaryOp(CExprOperator):
    """Base class for special n-ary operators."""

    __slots__ = ("args", )

    def __init__(self, args):
        self.args = [as_cexpr(arg) for arg in args]

    def ce_format(self, precision=None):
        # Format children
        args = [arg.ce_format(precision) for arg in self.args]

        # Apply parentheses
        for i in range(len(args)):
            if self.args[i].precedence >= self.precedence:
                args[i] = '(' + args[i] + ')'

        # Return combined string
        op = " " + self.op + " "
        s = args[0]
        for i in range(1, len(args)):
            s += op + args[i]
        return s

    def __eq__(self, other):
        return (isinstance(other, type(self)) and len(self.args) == len(other.args)
                and all(a == b for a, b in zip(self.args, other.args)))

    def flops(self):
        flops = len(self.args) - 1
        for arg in self.args:
            flops += arg.flops()
        return flops


# CExpr unary operators


class AddressOf(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.ADDRESSOF
    op = "&"


class SizeOf(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.SIZEOF
    op = "sizeof"


class Neg(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.NEG
    op = "-"


class Pos(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.POS
    op = "+"


class Not(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.NOT
    op = "!"


class BitNot(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.BIT_NOT
    op = "~"


class PreIncrement(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.PRE_INC
    sideeffect = True
    op = "++"


class PreDecrement(PrefixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.PRE_DEC
    sideeffect = True
    op = "--"


class PostIncrement(PostfixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.POST_INC
    sideeffect = True
    op = "++"


class PostDecrement(PostfixUnaryOp):
    __slots__ = ()
    precedence = PRECEDENCE.POST_DEC
    sideeffect = True
    op = "--"


# CExpr binary operators


class Add(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.ADD
    op = "+"


class Sub(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.SUB
    op = "-"


class Mul(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.MUL
    op = "*"


class Div(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.DIV
    op = "/"


class Mod(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.MOD
    op = "%"


class EQ(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.EQ
    op = "=="


class NE(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.NE
    op = "!="


class LT(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.LT
    op = "<"


class GT(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.GT
    op = ">"


class LE(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.LE
    op = "<="


class GE(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.GE
    op = ">="


class BitwiseAnd(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.BITAND
    op = "&"


class BitShiftR(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.BITSHIFT
    op = ">>"


class BitShiftL(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.BITSHIFT
    op = "<<"


class And(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.AND
    op = "&&"


class Or(BinOp):
    __slots__ = ()
    precedence = PRECEDENCE.OR
    op = "||"


class Sum(NaryOp):
    """Sum of any number of operands."""

    __slots__ = ()
    precedence = PRECEDENCE.ADD
    op = "+"


class Product(NaryOp):
    """Product of any number of operands."""

    __slots__ = ()
    precedence = PRECEDENCE.MUL
    op = "*"


class AssignOp(BinOp):
    """Base class for assignment operators."""

    __slots__ = ()
    precedence = PRECEDENCE.ASSIGN
    sideeffect = True

    def __init__(self, lhs, rhs):
        BinOp.__init__(self, as_cexpr_or_string_symbol(lhs), rhs)


class Assign(AssignOp):
    __slots__ = ()
    op = "="

    def flops(self):
        return super().flops() - 1


class AssignAdd(AssignOp):
    __slots__ = ()
    op = "+="


class AssignSub(AssignOp):
    __slots__ = ()
    op = "-="


class AssignMul(AssignOp):
    __slots__ = ()
    op = "*="


class AssignDiv(AssignOp):
    __slots__ = ()
    op = "/="

# CExpr operators


class FlattenedArray(object):
    """Syntax carrying object only, will get translated on __getitem__ to ArrayAccess."""

    __slots__ = ("array", "strides", "offset", "dims")

    def __init__(self, array, dummy=None, dims=None, strides=None, offset=None):
        assert dummy is None, "Please use keyword arguments for strides or dims."

        # Typecheck array argument
        if isinstance(array, ArrayDecl):
            self.array = array.symbol
        elif isinstance(array, Symbol):
            self.array = array
        else:
            assert isinstance(array, str)
            self.array = Symbol(array)

        # Allow expressions or literals as strides or dims and offset
        if strides is None:
            assert dims is not None, "Please provide either strides or dims."
            assert isinstance(dims, (list, tuple))
            dims = tuple(as_cexpr(i) for i in dims)
            self.dims = dims
            n = len(dims)
            literal_one = LiteralInt(1)
            strides = [literal_one] * n
            for i in range(n - 2, -1, -1):
                s = strides[i + 1]
                d = dims[i + 1]
                if d == literal_one:
                    strides[i] = s
                elif s == literal_one:
                    strides[i] = d
                else:
                    strides[i] = d * s
        else:
            self.dims = None
            assert isinstance(strides, (list, tuple))
            strides = tuple(as_cexpr(i) for i in strides)
        self.strides = strides
        self.offset = None if offset is None else as_cexpr(offset)

    def __getitem__(self, indices):
        if not isinstance(indices, (list, tuple)):
            indices = (indices, )
        n = len(indices)
        if n == 0:
            # Handle scalar case, allowing dims=() and indices=() for A[0]
            if len(self.strides) != 0:
                raise ValueError("Empty indices for nonscalar array.")
            flat = LiteralInt(0)
        else:
            i, s = (indices[0], self.strides[0])
            literal_one = LiteralInt(1)
            flat = (i if s == literal_one else s * i)
            if self.offset is not None:
                flat = self.offset + flat
            for i, s in zip(indices[1:n], self.strides[1:n]):
                flat = flat + (i if s == literal_one else s * i)
        # Delay applying ArrayAccess until we have all indices
        if n == len(self.strides):
            return ArrayAccess(self.array, flat)
        else:
            return FlattenedArray(self.array, strides=self.strides[n:], offset=flat)


class ArrayAccess(CExprOperator):
    __slots__ = ("array", "indices")
    precedence = PRECEDENCE.SUBSCRIPT

    def __init__(self, array, indices):
        # Typecheck array argument
        if isinstance(array, Symbol):
            self.array = array
        elif isinstance(array, ArrayDecl):
            self.array = array.symbol
        else:
            raise ValueError("Unexpected array type %s." % (type(array).__name__, ))

        # Allow expressions or literals as indices
        if not isinstance(indices, (list, tuple)):
            indices = (indices, )
        self.indices = tuple(as_cexpr_or_string_symbol(i) for i in indices)

        # Early error checking for negative array dimensions
        if any(isinstance(i, int) and i < 0 for i in self.indices):
            raise ValueError("Index value < 0.")

        # Additional dimension checks possible if we get an ArrayDecl instead of just a name
        if isinstance(array, ArrayDecl):
            if len(self.indices) != len(array.sizes):
                raise ValueError("Invalid number of indices.")
            ints = (int, LiteralInt)
            if any((isinstance(i, ints) and isinstance(d, ints) and int(i) >= int(d))
                   for i, d in zip(self.indices, array.sizes)):
                raise ValueError("Index value >= array dimension.")

    def __getitem__(self, indices):
        """Handle nested expr[i][j]."""
        if isinstance(indices, list):
            indices = tuple(indices)
        elif not isinstance(indices, tuple):
            indices = (indices, )
        return ArrayAccess(self.array, self.indices + indices)

    def ce_format(self, precision=None):
        s = self.array.ce_format(precision)
        for index in self.indices:
            s += "[" + index.ce_format(precision) + "]"
        return s

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.array == other.array
                and self.indices == other.indices)

    def __hash__(self):
        return hash(self.ce_format())

    def flops(self):
        return 0


class Conditional(CExprOperator):
    __slots__ = ("condition", "true", "false")
    precedence = PRECEDENCE.CONDITIONAL

    def __init__(self, condition, true, false):
        self.condition = as_cexpr(condition)
        self.true = as_cexpr(true)
        self.false = as_cexpr(false)

    def ce_format(self, precision=None):
        # Format children
        c = self.condition.ce_format(precision)
        t = self.true.ce_format(precision)
        f = self.false.ce_format(precision)

        # Apply parentheses
        if self.condition.precedence >= self.precedence:
            c = '(' + c + ')'
        if self.true.precedence >= self.precedence:
            t = '(' + t + ')'
        if self.false.precedence >= self.precedence:
            f = '(' + f + ')'

        # Return combined string
        return c + " ? " + t + " : " + f

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.condition == other.condition
                and self.true == other.true and self.false == other.false)

    def flops(self):
        raise NotImplementedError("Flop count is not implemented for conditionals")


class Call(CExprOperator):
    __slots__ = ("function", "arguments")
    precedence = PRECEDENCE.CALL
    sideeffect = True

    def __init__(self, function, arguments=None):
        self.function = as_cexpr_or_string_symbol(function)

        # Accept None, single, or multiple arguments; literals or CExprs
        if arguments is None:
            arguments = ()
        elif not isinstance(arguments, (tuple, list)):
            arguments = (arguments, )
        self.arguments = [as_cexpr(arg) for arg in arguments]

    def ce_format(self, precision=None):
        args = ", ".join(arg.ce_format(precision) for arg in self.arguments)
        return self.function.ce_format(precision) + "(" + args + ")"

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.function == other.function
                and self.arguments == other.arguments)

    def flops(self):
        return 1


def Sqrt(x):
    return Call("sqrt", x)


# Conversion function to expression nodes


def _is_zero_valued(values):
    if isinstance(values, (numbers.Integral, LiteralInt)):
        return int(values) == 0
    elif isinstance(values, (numbers.Number, LiteralFloat)):
        return float(values) == 0.0
    else:
        return numpy.count_nonzero(values) == 0


def as_cexpr(node):
    """Typechecks and wraps an object as a valid CExpr.

    Accepts CExpr nodes, treats int and float as literals, and treats a
    string as a symbol.

    """
    if isinstance(node, CExpr):
        return node
    elif isinstance(node, bool):
        return LiteralBool(node)
    elif isinstance(node, numbers.Integral):
        return LiteralInt(node)
    elif isinstance(node, numbers.Real):
        return LiteralFloat(node)
    elif isinstance(node, str):
        raise RuntimeError("Got string for CExpr, this is ambiguous: %s" % (node, ))
    else:
        raise RuntimeError("Unexpected CExpr type %s:\n%s" % (type(node), str(node)))


def as_cexpr_or_string_symbol(node):
    if isinstance(node, str):
        return Symbol(node)
    return as_cexpr(node)


def as_cexpr_or_literal(node):
    if isinstance(node, str):
        return LiteralString(node)
    return as_cexpr(node)


def as_symbol(symbol):
    if isinstance(symbol, str):
        symbol = Symbol(symbol)
    assert isinstance(symbol, Symbol)
    return symbol


def flattened_indices(indices, shape):
    """Return a flattened indexing expression.

    Given a tuple of indices and a shape tuple, return
    a CNode expression for flattened indexing into multidimensional
    array.

    Indices and shape entries can be int values, str symbol names, or
    CNode expressions.

    """
    n = len(shape)
    if n == 0:
        # Scalar
        return as_cexpr(0)
    elif n == 1:
        # Simple vector
        return as_cexpr(indices[0])
    else:
        # 2d or higher
        strides = [None] * (n - 2) + [shape[-1], 1]
        for i in range(n - 3, -1, -1):
            strides[i] = Mul(shape[i + 1], strides[i + 1])
        result = indices[-1]
        for i in range(n - 2, -1, -1):
            result = Add(Mul(strides[i], indices[i]), result)
        return result


# Base class for all statements


class CStatement(CNode):
    """Base class for all C statements.

    Subtypes do _not_ define a 'precedence' class attribute.

    """

    __slots__ = ()

    # True if statement contains its own scope, false by default to be
    # on the safe side
    is_scoped = False

    def cs_format(self, precision=None):
        """Return S: string | list(S) | Indented(S)."""
        raise NotImplementedError("Missing implementation of cs_format() in CStatement.")

    def __str__(self):
        try:
            s = self.cs_format()
        except Exception:
            logger.error("Error in CStatement string formatting.")
            raise
        return format_indented_lines(s)

    def flops(self):
        raise NotImplementedError()


# Statements


class VerbatimStatement(CStatement):
    """Wraps a source code string to be pasted verbatim into the source code."""

    __slots__ = ("codestring", )
    is_scoped = False

    def __init__(self, codestring):
        assert isinstance(codestring, str)
        self.codestring = codestring

    def cs_format(self, precision=None):
        return self.codestring

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.codestring == other.codestring)


class Statement(CStatement):
    """Make an expression into a statement."""

    __slots__ = ("expr", )
    is_scoped = False

    def __init__(self, expr):
        self.expr = as_cexpr(expr)

    def cs_format(self, precision=None):
        return self.expr.ce_format(precision) + ";"

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.expr == other.expr)

    def flops(self):
        # print(self.expr.rhs.flops())
        return self.expr.flops()


class StatementList(CStatement):
    """A simple sequence of statements. No new scopes are introduced."""

    __slots__ = ("statements", )

    def __init__(self, statements):
        self.statements = [as_cstatement(st) for st in statements]

    @property
    def is_scoped(self):
        return all(st.is_scoped for st in self.statements)

    def cs_format(self, precision=None):
        return [st.cs_format(precision) for st in self.statements]

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.statements == other.statements)

    def flops(self):
        flops = 0
        for statement in self.statements:
            flops += statement.flops()
        return flops


# Simple statements


class Break(CStatement):
    __slots__ = ()
    is_scoped = True

    def cs_format(self, precision=None):
        return "break;"

    def __eq__(self, other):
        return isinstance(other, type(self))

    def flops(self):
        return 0


class Continue(CStatement):
    __slots__ = ()
    is_scoped = True

    def cs_format(self, precision=None):
        return "continue;"

    def __eq__(self, other):
        return isinstance(other, type(self))

    def flops(self):
        return 0


class Return(CStatement):
    __slots__ = ("value", )
    is_scoped = True

    def __init__(self, value=None):
        if value is None:
            self.value = None
        else:
            self.value = as_cexpr(value)

    def cs_format(self, precision=None):
        if self.value is None:
            return "return;"
        else:
            return "return %s;" % (self.value.ce_format(precision), )

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.value == other.value)

    def flops(self):
        return 0


class Comment(CStatement):
    """Line comment(s) used for annotating the generated code with human readable remarks."""

    __slots__ = ("comment", )
    is_scoped = True

    def __init__(self, comment):
        assert isinstance(comment, str)
        self.comment = comment

    def cs_format(self, precision=None):
        lines = self.comment.strip().split("\n")
        return ["// " + line.strip() for line in lines]

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.comment == other.comment)

    def flops(self):
        return 0


def NoOp():
    return Comment("Do nothing")


def commented_code_list(code, comments):
    """Add comment to code list if the list is not empty."""
    if isinstance(code, CNode):
        code = [code]
    assert isinstance(code, list)
    if code:
        if not isinstance(comments, (list, tuple)):
            comments = [comments]
        comments = [Comment(c) for c in comments]
        code = comments + code
    return code


class Pragma(CStatement):
    """Pragma comments used for compiler-specific annotations."""

    __slots__ = ("comment", )
    is_scoped = True

    def __init__(self, comment):
        assert isinstance(comment, str)
        self.comment = comment

    def cs_format(self, precision=None):
        assert "\n" not in self.comment
        return "#pragma " + self.comment

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.comment == other.comment)

    def flops(self):
        return 0


# Type and variable declarations


class VariableDecl(CStatement):
    """Declare a variable, optionally define initial value."""

    __slots__ = ("typename", "symbol", "value")
    is_scoped = False

    def __init__(self, typename, symbol, value=None):

        # No type system yet, just using strings
        assert isinstance(typename, str)
        self.typename = typename

        # Allow Symbol or just a string
        self.symbol = as_symbol(symbol)

        if value is not None:
            value = as_cexpr(value)
        self.value = value

    def cs_format(self, precision=None):
        code = self.typename + " " + self.symbol.name
        if self.value is not None:
            code += " = " + self.value.ce_format(precision)
        return code + ";"

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.typename == other.typename
                and self.symbol == other.symbol and self.value == other.value)

    def flops(self):
        if self.value is not None:
            return self.value.flops()
        else:
            return 0


def leftover(size, padlen):
    """Return minimum integer to add to size to make it divisible by padlen."""
    return (padlen - (size % padlen)) % padlen


def pad_dim(dim, padlen):
    """Make dim divisible by padlen."""
    return ((dim + padlen - 1) // padlen) * padlen


def pad_innermost_dim(shape, padlen):
    """Make the last dimension in shape divisible by padlen."""
    if not shape:
        return ()
    shape = list(shape)
    if padlen:
        shape[-1] = pad_dim(shape[-1], padlen)
    return tuple(shape)


def build_1d_initializer_list(values, formatter, padlen=0, precision=None):
    """Return a list containing a single line formatted like '{ 0.0, 1.0, 2.0 }'."""
    if formatter == str:

        def formatter(x, p):
            return str(x)

    tokens = ["{ "]
    if numpy.product(values.shape) > 0:
        sep = ", "
        fvalues = [formatter(v, precision) for v in values]
        for v in fvalues[:-1]:
            tokens.append(v)
            tokens.append(sep)
        tokens.append(fvalues[-1])
        if padlen:
            # Add padding
            zero = formatter(values.dtype.type(0), precision)
            for i in range(leftover(len(values), padlen)):
                tokens.append(sep)
                tokens.append(zero)
    tokens += " }"
    return "".join(tokens)


def build_initializer_lists(values, sizes, level, formatter, padlen=0, precision=None):
    """Return a list of lines with initializer lists for a multidimensional array.

    Example output::

        { { 0.0, 0.1 },
          { 1.0, 1.1 } }

    """
    if formatter == str:

        def formatter(x, p):
            return str(x)

    values = numpy.asarray(values)
    assert numpy.product(values.shape) == numpy.product(sizes)
    assert len(sizes) > 0
    assert len(values.shape) > 0
    assert len(sizes) == len(values.shape)
    assert numpy.all(values.shape == sizes)

    r = len(sizes)
    assert r > 0
    if r == 1:
        return [build_1d_initializer_list(values, formatter, padlen=padlen, precision=precision)]
    else:
        # Render all sublists
        parts = []
        for val in values:
            sublist = build_initializer_lists(
                val, sizes[1:], level + 1, formatter, padlen=padlen, precision=precision)
            parts.append(sublist)
        # Add comma after last line in each part except the last one
        for part in parts[:-1]:
            part[-1] += ","
        # Collect all lines in flat list
        lines = []
        for part in parts:
            lines.extend(part)
        # Enclose lines in '{ ' and ' }' and indent lines in between
        lines[0] = "{ " + lines[0]
        for i in range(1, len(lines)):
            lines[i] = "  " + lines[i]
        lines[-1] += " }"
        return lines


class ArrayDecl(CStatement):
    """A declaration or definition of an array.

    Note that just setting values=0 is sufficient to initialize the
    entire array to zero.

    Otherwise use nested lists of lists to represent multidimensional
    array values to initialize to.

    """

    __slots__ = ("typename", "symbol", "sizes", "padlen", "values")
    is_scoped = False

    def __init__(self, typename, symbol, sizes=None, values=None, padlen=0):
        assert isinstance(typename, str)
        self.typename = typename

        if isinstance(symbol, FlattenedArray):
            if sizes is None:
                assert symbol.dims is not None
                sizes = symbol.dims
            elif symbol.dims is not None:
                assert symbol.dims == sizes
            self.symbol = symbol.array
        else:
            self.symbol = as_symbol(symbol)

        if isinstance(sizes, int):
            sizes = (sizes, )
        self.sizes = tuple(sizes)

        # NB! No type checking, assuming nested lists of literal values. Not applying as_cexpr.
        if isinstance(values, (list, tuple)):
            self.values = numpy.asarray(values)
        else:
            self.values = values

        self.padlen = padlen

    def cs_format(self, precision=None):
        if not all(self.sizes):
            raise RuntimeError(f"Detected an array {self.symbol} dimension of zero. This is not valid in C.")

        # Pad innermost array dimension
        sizes = pad_innermost_dim(self.sizes, self.padlen)

        # Add brackets
        brackets = ''.join("[%d]" % n for n in sizes)

        # Join declaration
        decl = self.typename + " " + self.symbol.name + brackets

        if self.values is None:
            # Undefined initial values
            return decl + ";"
        elif _is_zero_valued(self.values):
            # Zero initial values
            # (NB! C style zero initialization, not sure about other target languages)
            nb = len(sizes)
            lbr = "{" * nb
            rbr = "}" * nb
            return f"{decl} = {lbr} 0 {rbr};"
        else:
            # Construct initializer lists for arbitrary multidimensional array values
            if self.values.dtype.kind == "f":
                formatter = format_float
            elif self.values.dtype.kind == "i":
                formatter = format_int
            elif self.values.dtype == numpy.bool_:
                def format_bool(x, precision=None):
                    return "true" if x is True else "false"
                formatter = format_bool
            else:
                formatter = format_value
            initializer_lists = build_initializer_lists(
                self.values, self.sizes, 0, formatter, padlen=self.padlen, precision=precision)
            if len(initializer_lists) == 1:
                return decl + " = " + initializer_lists[0] + ";"
            else:
                initializer_lists[-1] += ";"  # Close statement on final line
                return (decl + " =", Indented(initializer_lists))

    def __eq__(self, other):
        attributes = ("typename", "symbol", "sizes", "padlen", "values")
        return (isinstance(other, type(self))
                and all(getattr(self, name) == getattr(self, name) for name in attributes))

    def flops(self):
        return 0


# Scoped statements


class Scope(CStatement):
    __slots__ = ("body", )
    is_scoped = True

    def __init__(self, body):
        self.body = as_cstatement(body)

    def cs_format(self, precision=None):
        return ("{", Indented(self.body.cs_format(precision)), "}")

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.body == other.body)

    def flops(self):
        return 0


def _is_simple_if_body(body):
    if isinstance(body, StatementList):
        if len(body.statements) > 1:
            return False
        body, = body.statements
    return isinstance(body, (Return, AssignOp, Break, Continue))


class If(CStatement):
    __slots__ = ("condition", "body")
    is_scoped = True

    def __init__(self, condition, body):
        self.condition = as_cexpr(condition)
        self.body = as_cstatement(body)

    def cs_format(self, precision=None):
        statement = "if (" + self.condition.ce_format(precision) + ")"
        body_fmt = Indented(self.body.cs_format(precision))
        if _is_simple_if_body(self.body):
            return (statement, body_fmt)
        else:
            return (statement, "{", body_fmt, "}")

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.condition == other.condition
                and self.body == other.body)


class ElseIf(CStatement):
    __slots__ = ("condition", "body")
    is_scoped = True

    def __init__(self, condition, body):
        self.condition = as_cexpr(condition)
        self.body = as_cstatement(body)

    def cs_format(self, precision=None):
        statement = "else if (" + self.condition.ce_format(precision) + ")"
        body_fmt = Indented(self.body.cs_format(precision))
        if _is_simple_if_body(self.body):
            return (statement, body_fmt)
        else:
            return (statement, "{", body_fmt, "}")

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.condition == other.condition
                and self.body == other.body)


class Else(CStatement):
    __slots__ = ("body", )
    is_scoped = True

    def __init__(self, body):
        self.body = as_cstatement(body)

    def cs_format(self, precision=None):
        statement = "else"
        body_fmt = Indented(self.body.cs_format(precision))
        if _is_simple_if_body(self.body):
            return (statement, body_fmt)
        else:
            return (statement, "{", body_fmt, "}")

    def __eq__(self, other):
        return (isinstance(other, type(self)) and self.body == other.body)


def is_simple_inner_loop(code):
    if isinstance(code, ForRange) and is_simple_inner_loop(code.body):
        return True
    if isinstance(code, Statement) and isinstance(code.expr, AssignOp):
        return True
    return False


class Switch(CStatement):
    __slots__ = ("arg", "cases", "default", "autobreak", "autoscope")
    is_scoped = True

    def __init__(self, arg, cases, default=None, autobreak=True, autoscope=True):
        self.arg = as_cexpr_or_string_symbol(arg)
        self.cases = [(as_cexpr(value), as_cstatement(body)) for value, body in cases]
        if default is not None:
            default = as_cstatement(default)
            defcase = [(None, default)]
        else:
            defcase = []
        self.default = default
        # If this is a switch where every case returns, scopes or breaks are never needed
        if all(isinstance(case[1], Return) for case in self.cases + defcase):
            autobreak = False
            autoscope = False
        if all(case[1].is_scoped for case in self.cases + defcase):
            autoscope = False
        assert autobreak in (True, False)
        assert autoscope in (True, False)
        self.autobreak = autobreak
        self.autoscope = autoscope

    def cs_format(self, precision=None):
        cases = []
        for case in self.cases:
            caseheader = "case " + case[0].ce_format(precision) + ":"
            casebody = case[1].cs_format(precision)
            if self.autoscope:
                casebody = ("{", Indented(casebody), "}")
            if self.autobreak:
                casebody = (casebody, "break;")
            cases.extend([caseheader, Indented(casebody)])

        if self.default is not None:
            caseheader = "default:"
            casebody = self.default.cs_format(precision)
            if self.autoscope:
                casebody = ("{", Indented(casebody), "}")
            cases.extend([caseheader, Indented(casebody)])

        return ("switch (" + self.arg.ce_format(precision) + ")", "{", cases, "}")

    def __eq__(self, other):
        attributes = ("arg", "cases", "default", "autobreak", "autoscope")
        return (isinstance(other, type(self))
                and all(getattr(self, name) == getattr(self, name) for name in attributes))


class ForRange(CStatement):
    """Slightly higher-level for loop assuming incrementing an index over a range."""

    __slots__ = ("index", "begin", "end", "body", "index_type")
    is_scoped = True

    def __init__(self, index, begin, end, body, index_type="int"):
        self.index = as_cexpr_or_string_symbol(index)
        self.begin = as_cexpr(begin)
        self.end = as_cexpr(end)
        self.body = as_cstatement(body)
        self.index_type = index_type

    def cs_format(self, precision=None):
        indextype = self.index_type
        index = self.index.ce_format(precision)
        begin = self.begin.ce_format(precision)
        end = self.end.ce_format(precision)

        init = indextype + " " + index + " = " + begin
        check = index + " < " + end
        update = "++" + index

        prelude = "for (" + init + "; " + check + "; " + update + ")"
        body = Indented(self.body.cs_format(precision))

        # Reduce size of code with lots of simple loops by dropping {} in obviously safe cases
        if is_simple_inner_loop(self.body):
            code = (prelude, body)
        else:
            code = (prelude, "{", body, "}")

        return code

    def __eq__(self, other):
        attributes = ("index", "begin", "end", "body", "index_type")
        return (isinstance(other, type(self))
                and all(getattr(self, name) == getattr(self, name) for name in attributes))

    def flops(self):
        return (self.end.value - self.begin.value) * self.body.flops()


# Conversion function to statement nodes


def as_cstatement(node):
    """Perform type checking on node and wrap in a suitable statement type if necessary."""
    if isinstance(node, StatementList) and len(node.statements) == 1:
        # Cleans up the expression tree a bit
        return node.statements[0]
    elif isinstance(node, CStatement):
        # No-op
        return node
    elif isinstance(node, CExprOperator):
        if node.sideeffect:
            # Special case for using assignment expressions as statements
            return Statement(node)
        else:
            raise RuntimeError("Trying to create a statement of CExprOperator type %s:\n%s" %
                               (type(node), str(node)))
    elif isinstance(node, list):
        # Convenience case for list of statements
        if len(node) == 1:
            # Cleans up the expression tree a bit
            return as_cstatement(node[0])
        else:
            return StatementList(node)
    elif isinstance(node, str):
        # Backdoor for flexibility in code generation to allow verbatim pasted statements
        return VerbatimStatement(node)
    else:
        raise RuntimeError("Unexpected CStatement type %s:\n%s" % (type(node), str(node)))



# Code generation format strings for UFC (Unified Form-assembly Code)
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2018.

factory = """
// Code for custom element {factory_name}

{value_shape_init}
{wcoeffs_init}
{npts_init}
{ndofs_init}
{x_init}
{M_init}

ufcx_basix_custom_finite_element {factory_name} =
{{
  .cell_type = {cell_type},
  .value_shape_length = {value_shape_length},
  .value_shape = {value_shape},
  .wcoeffs_rows = {wcoeffs_rows},
  .wcoeffs_cols = {wcoeffs_cols},
  .wcoeffs = {wcoeffs},
  .npts = {npts},
  .ndofs = {ndofs},
  .x = {x},
  .M = {M},
  .map_type = {map_type},
  .sobolev_space = {sobolev_space},
  .discontinuous = {discontinuous},
  .highest_complete_degree = {highest_complete_degree},
  .interpolation_nderivs = {interpolation_nderivs},
  .highest_degree = {highest_degree}
}};

// End of code for custom element {factory_name}
"""



# Copyright (C) 2009-2017 Anders Logg, Martin Sandve Alnæs, Marie E. Rognes,
# Kristian B. Oelgaard, and others
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Compiler stage 4: Code generation.

This module implements the generation of C code for the body of each
UFC function from an intermediate representation (IR).

"""

import logging
import typing

from ffcx.codegeneration.dofmap import generator as dofmap_generator
from ffcx.codegeneration.expressions import generator as expression_generator
from ffcx.codegeneration.finite_element import \
    generator as finite_element_generator
from ffcx.codegeneration.form import generator as form_generator
from ffcx.codegeneration.integrals import generator as integral_generator

logger = logging.getLogger("ffcx")


class CodeBlocks(typing.NamedTuple):
    """
    Storage of code blocks of the form (declaration, implementation).

    Blocks for elements, dofmaps, integrals, forms and expressions is stored
    """

    elements: typing.List[typing.Tuple[str, str]]
    dofmaps: typing.List[typing.Tuple[str, str]]
    integrals: typing.List[typing.Tuple[str, str]]
    forms: typing.List[typing.Tuple[str, str]]
    expressions: typing.List[typing.Tuple[str, str]]


def generate_code(ir, options) -> CodeBlocks:
    """Generate code blocks from intermediate representation."""
    logger.info(79 * "*")
    logger.info("Compiler stage 3: Generating code")
    logger.info(79 * "*")

    # Generate code for finite_elements
    code_finite_elements = [finite_element_generator(element_ir, options) for element_ir in ir.elements]
    code_dofmaps = [dofmap_generator(dofmap_ir, options) for dofmap_ir in ir.dofmaps]
    code_integrals = [integral_generator(integral_ir, options) for integral_ir in ir.integrals]
    code_forms = [form_generator(form_ir, options) for form_ir in ir.forms]
    code_expressions = [expression_generator(expression_ir, options) for expression_ir in ir.expressions]
    return CodeBlocks(elements=code_finite_elements, dofmaps=code_dofmaps,
                      integrals=code_integrals, forms=code_forms, expressions=code_expressions)



# Copyright (C) 2005-2020 Anders Logg, Michal Habera, Jack S. Hale
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import functools
import json
import logging
import os
import os.path
import pprint
from pathlib import Path
from typing import Any, Dict, Optional

logger = logging.getLogger("ffcx")

FFCX_DEFAULT_OPTIONS = {
    "epsilon":
        (1e-14, "Machine precision, used for dropping zero terms in tables"),
    "scalar_type":
        ("double", """Scalar type used in generated code. Any of real or complex C floating-point types, e.g.
                      float, double, float _Complex, double _Complex, ..."""),
    "tabulate_tensor_void":
        (False, "True to generate empty tabulation kernels."),
    "table_rtol":
        (1e-6, "Relative precision to use when comparing finite element table values for table reuse."),
    "table_atol":
        (1e-9, "Absolute precision to use when comparing finite element table values for reuse."),
    "assume_aligned":
        (-1, """Assumes alignment (in bytes) of pointers to tabulated tensor, coefficients and constants array.
               This value must be compatible with alignment of data structures allocated outside FFC.
               (-1 means no alignment assumed, safe option)"""),
    "padlen":
        (1, "Pads every declared array in tabulation kernel such that its last dimension is divisible by given value."),
    "verbosity":
        (30, "Logger verbosity. Follows standard logging library levels, i.e. INFO=20, DEBUG=10, etc.")
}


@functools.lru_cache(maxsize=None)
def _load_options():
    """Load options from JSON files."""
    user_config_file = os.getenv("XDG_CONFIG_HOME", default=Path.home().joinpath(".config")) \
        / Path("ffcx", "ffcx_options.json")
    try:
        with open(user_config_file) as f:
            user_options = json.load(f)
    except FileNotFoundError:
        user_options = {}

    pwd_config_file = Path.cwd().joinpath("ffcx_options.json")
    try:
        with open(pwd_config_file) as f:
            pwd_options = json.load(f)
    except FileNotFoundError:
        pwd_options = {}

    return (user_options, pwd_options)


def get_options(priority_options: Optional[dict] = None) -> dict:
    """Return (a copy of) the merged option values for FFCX.

    Options
    ----------
      priority_options:
        take priority over all other option values (see notes)

    Returns
    -------
      dict: merged option values

    Notes
    -----
    This function sets the log level from the merged option values prior to
    returning.

    The `ffcx_options.json` files are cached on the first call. Subsequent
    calls to this function use this cache.

    Priority ordering of options from highest to lowest is:

    -  **priority_options** (API and command line options)
    -  **$PWD/ffcx_options.json** (local options)
    -  **$XDG_CONFIG_HOME/ffcx/ffcx_options.json** (user options)
    -  **FFCX_DEFAULT_OPTIONS** in `ffcx.options`

    `XDG_CONFIG_HOME` is `~/.config/` if the environment variable is not set.

    Example `ffcx_options.json` file:

      { "assume_aligned": 32, "epsilon": 1e-7 }

    """
    options: Dict[str, Any] = {}

    for opt, (value, _) in FFCX_DEFAULT_OPTIONS.items():
        options[opt] = value

    # NOTE: _load_options uses functools.lru_cache
    user_options, pwd_options = _load_options()

    options.update(user_options)
    options.update(pwd_options)
    if priority_options is not None:
        options.update(priority_options)

    logger.setLevel(options["verbosity"])

    logger.info("Final option values")
    logger.info(pprint.pformat(options))

    return options



import os
import sys

import pytest

demo_dir = os.path.dirname(os.path.realpath(__file__))

ufl_files = []
for file in os.listdir(demo_dir):
    if file.endswith(".py") and not file == "test_demos.py":
        ufl_files.append(file[:-3])


@pytest.mark.parametrize("file", ufl_files)
def test_demo(file):
    if file in [
        "MixedGradient", "TraceElement",  # HDiv Trace
        "MixedElasticity",  # VectorElement of BDM
        "RestrictedElement",
        "_TensorProductElement"
    ]:
        # Skip demos that use elements not yet implemented in Basix
        pytest.skip()

    extra_flags = "-Wunused-variable -Werror -fPIC "
    assert os.system(f"cd {demo_dir} && ffcx {file}.py") == 0
    assert os.system(f"cd {demo_dir} && "
                     "CPATH=../ffcx/codegeneration/ "
                     f"gcc -I/usr/include/python{sys.version_info.major}.{sys.version_info.minor} {extra_flags}"
                     f"-shared {file}.c -o {file}.so") == 0



#!/usr/bin/env python
# Copyright (C) 2017-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ffcx.main import main

if __name__ == "__main__":
    import sys
    sys.exit(main())



# Copyright (C) 2021 Igor A. Baratta
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from typing import Optional

import ffcx.options
import ufl
from ffcx.analysis import analyze_ufl_objects
from ffcx.codegeneration.backend import FFCXBackend
from ffcx.codegeneration.integrals import IntegralGenerator
from ffcx.ir.representation import compute_ir


def count_flops(form: ufl.Form, options: Optional[dict] = {}):
    """Return a list with the number of flops for each kernel in the Form."""
    options = ffcx.options.get_options(options)
    assert isinstance(form, ufl.Form)
    analysis = analyze_ufl_objects([form], options)
    ir = compute_ir(analysis, {}, "flops", options, False)

    flops = []

    for integral_ir in ir.integrals:
        # Create FFCx C backend
        backend = FFCXBackend(integral_ir, options)
        # Configure kernel generator
        ig = IntegralGenerator(integral_ir, backend)
        # Generate code ast for the tabulate_tensor body
        ast = ig.generate()
        flops.append(ast.flops())

    return flops



# Copyright (C) 2011 Garth N. Wells
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# This example demonstrates how to create vectors component-wise
from ufl import (Coefficient, TestFunction, VectorElement, as_vector, dot, dx,
                 tetrahedron)

element = VectorElement("Lagrange", tetrahedron, 1)

v = TestFunction(element)
f = Coefficient(element)

# Create vector
v0 = as_vector([v[0], v[1], 0.0])

# Use created vector in linear form
L = dot(f, v0) * dx



# Copyright (C) 2009-2022 Anders Logg, Martin Sandve Alnæs, Matthew Scroggs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# Note: Much of the code in this file is a direct translation
# from the old implementation in FFC, although some improvements
# have been made to the generated code.

import logging

import ffcx.codegeneration.basix_custom_element_template as ufcx_basix_custom_finite_element
import ffcx.codegeneration.finite_element_template as ufcx_finite_element
import ufl

logger = logging.getLogger("ffcx")
index_type = "int"


def generator(ir, options):
    """Generate UFC code for a finite element."""
    logger.info("Generating code for finite element:")
    logger.info(f"--- family: {ir.family}")
    logger.info(f"--- degree: {ir.degree}")
    logger.info(f"--- value shape: {ir.value_shape}")
    logger.info(f"--- name: {ir.name}")

    d = {}
    d["factory_name"] = ir.name
    d["signature"] = f"\"{ir.signature}\""
    d["geometric_dimension"] = ir.geometric_dimension
    d["topological_dimension"] = ir.topological_dimension
    d["cell_shape"] = ir.cell_shape
    d["element_type"] = ir.element_type
    d["space_dimension"] = ir.space_dimension
    d["value_rank"] = len(ir.value_shape)
    d["value_size"] = ufl.product(ir.value_shape)
    d["reference_value_rank"] = len(ir.reference_value_shape)
    d["reference_value_size"] = ufl.product(ir.reference_value_shape)
    d["degree"] = ir.degree
    d["family"] = f"\"{ir.family}\""
    d["num_sub_elements"] = ir.num_sub_elements
    d["block_size"] = ir.block_size
    d["discontinuous"] = "true" if ir.discontinuous else "false"

    if ir.lagrange_variant is None:
        d["lagrange_variant"] = -1
    else:
        d["lagrange_variant"] = int(ir.lagrange_variant)

    if ir.dpc_variant is None:
        d["dpc_variant"] = -1
    else:
        d["dpc_variant"] = int(ir.dpc_variant)

    if ir.basix_family is None:
        d["basix_family"] = -1
    else:
        d["basix_family"] = int(ir.basix_family)
    if ir.basix_cell is None:
        d["basix_cell"] = -1
    else:
        d["basix_cell"] = int(ir.basix_cell)

    import ffcx.codegeneration.C.cnodes as L

    if len(ir.value_shape) > 0:
        d["value_shape"] = f"value_shape_{ir.name}"
        d["value_shape_init"] = L.ArrayDecl(
            "int", f"value_shape_{ir.name}", values=ir.value_shape, sizes=len(ir.value_shape))
    else:
        d["value_shape"] = "NULL"
        d["value_shape_init"] = ""

    if len(ir.reference_value_shape) > 0:
        d["reference_value_shape"] = f"reference_value_shape_{ir.name}"
        d["reference_value_shape_init"] = L.ArrayDecl(
            "int", f"reference_value_shape_{ir.name}",
            values=ir.reference_value_shape, sizes=len(ir.reference_value_shape))
    else:
        d["reference_value_shape"] = "NULL"
        d["reference_value_shape_init"] = ""

    if len(ir.sub_elements) > 0:
        d["sub_elements"] = f"sub_elements_{ir.name}"
        d["sub_elements_init"] = L.ArrayDecl(
            "ufcx_finite_element*", f"sub_elements_{ir.name}",
            values=[L.AddressOf(L.Symbol(el)) for el in ir.sub_elements], sizes=len(ir.sub_elements))
    else:
        d["sub_elements"] = "NULL"
        d["sub_elements_init"] = ""

    if ir.custom_element is not None:
        d["custom_element"] = f"&custom_element_{ir.name}"
        d["custom_element_init"] = generate_custom_element(f"custom_element_{ir.name}", ir.custom_element)
    else:
        d["custom_element"] = "NULL"
        d["custom_element_init"] = ""

    # Check that no keys are redundant or have been missed
    from string import Formatter
    fieldnames = [
        fname for _, fname, _, _ in Formatter().parse(ufcx_finite_element.factory) if fname
    ]
    assert set(fieldnames) == set(
        d.keys()), "Mismatch between keys in template and in formatting dict"

    # Format implementation code
    implementation = ufcx_finite_element.factory.format_map(d)

    # Format declaration
    declaration = ufcx_finite_element.declaration.format(factory_name=ir.name)

    return declaration, implementation


def generate_custom_element(name, ir):
    d = {}
    d["factory_name"] = name
    d["cell_type"] = int(ir.cell_type)
    d["map_type"] = int(ir.map_type)
    d["sobolev_space"] = int(ir.sobolev_space)
    d["highest_complete_degree"] = ir.highest_complete_degree
    d["highest_degree"] = ir.highest_degree
    d["discontinuous"] = "true" if ir.discontinuous else "false"
    d["interpolation_nderivs"] = ir.interpolation_nderivs

    import ffcx.codegeneration.C.cnodes as L

    d["value_shape_length"] = len(ir.value_shape)
    if len(ir.value_shape) > 0:
        d["value_shape"] = f"value_shape_{name}"
        d["value_shape_init"] = L.ArrayDecl(
            "int", f"value_shape_{name}", values=ir.value_shape, sizes=len(ir.value_shape))
    else:
        d["value_shape"] = "NULL"
        d["value_shape_init"] = ""

    d["wcoeffs_rows"] = ir.wcoeffs.shape[0]
    d["wcoeffs_cols"] = ir.wcoeffs.shape[1]
    d["wcoeffs"] = f"wcoeffs_{name}"
    d["wcoeffs_init"] = f"double wcoeffs_{name}[{ir.wcoeffs.shape[0] * ir.wcoeffs.shape[1]}] = "
    d["wcoeffs_init"] += "{" + ",".join([f" {i}" for row in ir.wcoeffs for i in row]) + "};"

    npts = []
    x = []
    for entity in ir.x:
        for points in entity:
            npts.append(points.shape[0])
            for row in points:
                for i in row:
                    x.append(i)
    d["npts"] = f"npts_{name}"
    d["npts_init"] = f"int npts_{name}[{len(npts)}] = "
    d["npts_init"] += "{" + ",".join([f" {i}" for i in npts]) + "};"
    d["x"] = f"x_{name}"
    d["x_init"] = f"double x_{name}[{len(x)}] = "
    d["x_init"] += "{" + ",".join([f" {i}" for i in x]) + "};"
    ndofs = []
    M = []
    for entity in ir.M:
        for mat4d in entity:
            ndofs.append(mat4d.shape[0])
            for mat3d in mat4d:
                for mat2d in mat3d:
                    for row in mat2d:
                        for i in row:
                            M.append(i)

    d["ndofs"] = f"ndofs_{name}"
    d["ndofs_init"] = f"int ndofs_{name}[{len(ndofs)}] = "
    d["ndofs_init"] += "{" + ",".join([f" {i}" for i in ndofs]) + "};"
    d["M"] = f"M_{name}"
    d["M_init"] = f"double M_{name}[{len(M)}] = "
    d["M_init"] += "{" + ",".join([f" {i}" for i in M]) + "};"

    # Check that no keys are redundant or have been missed
    from string import Formatter
    fieldnames = [
        fname for _, fname, _, _ in Formatter().parse(ufcx_basix_custom_finite_element.factory) if fname
    ]
    assert set(fieldnames) == set(
        d.keys()), "Mismatch between keys in template and in formatting dict"

    # Format implementation code
    implementation = ufcx_basix_custom_finite_element.factory.format_map(d)

    return implementation



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

"""Algorithms for the analysis phase of the form compilation."""



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Algorithms for value numbering within computational graphs."""

import logging

import ufl
from ffcx.ir.analysis.indexing import (map_component_tensor_arg_components,
                                       map_indexed_arg_components)
from ffcx.ir.analysis.modified_terminals import analyse_modified_terminal

logger = logging.getLogger("ffcx")


class ValueNumberer(object):
    """Maps scalar components to unique values.

    An algorithm to map the scalar components of an expression node to unique value numbers,
    with fallthrough for types that can be mapped to the value numbers
    of their operands.
    """

    def __init__(self, G):
        self.symbol_count = 0
        self.G = G
        self.V_symbols = []
        self.call_lookup = {ufl.classes.Expr: self.expr,
                            ufl.classes.Argument: self.form_argument,
                            ufl.classes.Coefficient: self.form_argument,
                            ufl.classes.Grad: self._modified_terminal,
                            ufl.classes.ReferenceGrad: self._modified_terminal,
                            ufl.classes.FacetAvg: self._modified_terminal,
                            ufl.classes.CellAvg: self._modified_terminal,
                            ufl.classes.Restricted: self._modified_terminal,
                            ufl.classes.ReferenceValue: self._modified_terminal,
                            ufl.classes.Indexed: self.indexed,
                            ufl.classes.ComponentTensor: self.component_tensor,
                            ufl.classes.ListTensor: self.list_tensor,
                            ufl.classes.Variable: self.variable}

    def new_symbols(self, n):
        """Generate new symbols with a running counter."""
        begin = self.symbol_count
        end = begin + n
        self.symbol_count = end
        return list(range(begin, end))

    def new_symbol(self):
        """Generate new symbol with a running counter."""
        begin = self.symbol_count
        self.symbol_count += 1
        return begin

    def get_node_symbols(self, expr):
        idx = [i for i, v in self.G.nodes.items() if v['expression'] == expr][0]
        return self.V_symbols[idx]

    def compute_symbols(self):
        for i, v in self.G.nodes.items():
            expr = v['expression']
            symbol = None
            # First look for exact type match
            f = self.call_lookup.get(type(expr), False)
            if f:
                symbol = f(expr)
            else:
                # Look for parent class types instead
                for k in self.call_lookup.keys():
                    if isinstance(expr, k):
                        symbol = self.call_lookup[k](expr)
                        break

            if symbol is None:
                # Nothing found
                raise RuntimeError("Not expecting type %s here." % type(expr))

            self.V_symbols.append(symbol)

        return self.V_symbols

    def expr(self, v):
        """Create new symbols for expressions that represent new values."""
        n = ufl.product(v.ufl_shape + v.ufl_index_dimensions)
        return self.new_symbols(n)

    def form_argument(self, v):
        """Create new symbols for expressions that represent new values."""
        symmetry = v.ufl_function_space().ufl_element().symmetry()

        if symmetry:
            # Build symbols with symmetric components skipped
            symbols = []
            mapped_symbols = {}
            for c in ufl.permutation.compute_indices(v.ufl_shape):
                # Build mapped component mc with symmetries from element considered
                mc = symmetry.get(c, c)

                # Get existing symbol or create new and store with mapped component mc as key
                s = mapped_symbols.get(mc)
                if s is None:
                    s = self.new_symbol()
                    mapped_symbols[mc] = s
                symbols.append(s)
        else:
            n = ufl.product(v.ufl_shape + v.ufl_index_dimensions)
            symbols = self.new_symbols(n)

        return symbols

    # Handle modified terminals with element symmetries and second derivative symmetries!
    # terminals are implemented separately, or maybe they don't need to be?

    def _modified_terminal(self, v):
        """Handle modified terminal.

        Modifiers:
        ---------
        terminal           - the underlying Terminal object
        global_derivatives - tuple of ints, each meaning derivative in that global direction
        local_derivatives  - tuple of ints, each meaning derivative in that local direction
        reference_value    - bool, whether this is represented in reference frame
        averaged           - None, 'facet' or 'cell'
        restriction        - None, '+' or '-'
        component          - tuple of ints, the global component of the Terminal
        flat_component     - single int, flattened local component of the Terminal, considering symmetry

        """
        # (1) mt.terminal.ufl_shape defines a core indexing space UNLESS mt.reference_value,
        #     in which case the reference value shape of the element must be used.
        # (2) mt.terminal.ufl_element().symmetry() defines core symmetries
        # (3) averaging and restrictions define distinct symbols, no additional symmetries
        # (4) two or more grad/reference_grad defines distinct symbols with additional symmetries

        # v is not necessary scalar here, indexing in (0,...,0) picks the first scalar component
        # to analyse, which should be sufficient to get the base shape and derivatives
        if v.ufl_shape:
            mt = analyse_modified_terminal(v[(0, ) * len(v.ufl_shape)])
        else:
            mt = analyse_modified_terminal(v)

        # Get derivatives
        num_ld = len(mt.local_derivatives)
        num_gd = len(mt.global_derivatives)
        assert not (num_ld and num_gd)
        if num_ld:
            domain = ufl.domain.extract_unique_domain(mt.terminal)
            tdim = domain.topological_dimension()
            d_components = ufl.permutation.compute_indices((tdim, ) * num_ld)
        elif num_gd:
            domain = ufl.domain.extract_unique_domiain(mt.terminal)
            gdim = domain.geometric_dimension()
            d_components = ufl.permutation.compute_indices((gdim, ) * num_gd)
        else:
            d_components = [()]

        # Get base shape without the derivative axes
        base_components = ufl.permutation.compute_indices(mt.base_shape)

        # Build symbols with symmetric components and derivatives skipped
        symbols = []
        mapped_symbols = {}
        for bc in base_components:
            for dc in d_components:
                # Build mapped component mc with symmetries from element
                # and derivatives combined
                mbc = mt.base_symmetry.get(bc, bc)
                mdc = tuple(sorted(dc))
                mc = mbc + mdc

                # Get existing symbol or create new and store with
                # mapped component mc as key
                s = mapped_symbols.get(mc)
                if s is None:
                    s = self.new_symbol()
                    mapped_symbols[mc] = s
                symbols.append(s)

        # Consistency check before returning symbols
        assert not v.ufl_free_indices
        if ufl.product(v.ufl_shape) != len(symbols):
            raise RuntimeError("Internal error in value numbering.")
        return symbols

    # indexed is implemented as a fall-through operation
    def indexed(self, Aii):
        # Reuse symbols of arg A for Aii
        A = Aii.ufl_operands[0]

        # Get symbols of argument A
        A_symbols = self.get_node_symbols(A)

        # Map A_symbols to Aii_symbols
        d = map_indexed_arg_components(Aii)
        symbols = [A_symbols[k] for k in d]
        return symbols

    def component_tensor(self, A):
        # Reuse symbols of arg Aii for A
        Aii = A.ufl_operands[0]

        # Get symbols of argument Aii
        Aii_symbols = self.get_node_symbols(Aii)

        # Map A_symbols to Aii_symbols
        d = map_component_tensor_arg_components(A)
        symbols = [Aii_symbols[k] for k in d]
        return symbols

    def list_tensor(self, v):
        symbols = []
        for row in v.ufl_operands:
            symbols.extend(self.get_node_symbols(row))
        return symbols

    def variable(self, v):
        """Direct reuse of all symbols."""
        return self.get_node_symbols(v.ufl_operands[0])



from ufl import (TensorElement, TestFunction, TrialFunction, dx, grad, inner,
                 triangle)

P1 = TensorElement("CG", triangle, 1, (2, 2), symmetry={(1, 0): (0, 1)})

u = TrialFunction(P1)
v = TestFunction(P1)

a = inner(grad(u), grad(v)) * dx



# Code generation format strings for UFC (Unified Form-assembly Code)
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2018.

declaration = """
extern ufcx_finite_element {factory_name};
"""

factory = """
// Code for element {factory_name}

{value_shape_init}
{reference_value_shape_init}
{sub_elements_init}
{custom_element_init}

ufcx_finite_element {factory_name} =
{{
  .signature = {signature},
  .cell_shape = {cell_shape},
  .element_type = {element_type},
  .topological_dimension = {topological_dimension},
  .geometric_dimension = {geometric_dimension},
  .space_dimension = {space_dimension},
  .value_rank = {value_rank},
  .value_shape = {value_shape},
  .value_size = {value_size},
  .reference_value_rank = {reference_value_rank},
  .reference_value_shape = {reference_value_shape},
  .reference_value_size = {reference_value_size},
  .degree = {degree},
  .block_size = {block_size},
  .family = {family},
  .basix_family = {basix_family},
  .basix_cell = {basix_cell},
  .discontinuous = {discontinuous},
  .lagrange_variant = {lagrange_variant},
  .dpc_variant = {dpc_variant},
  .num_sub_elements = {num_sub_elements},
  .sub_elements = {sub_elements},
  .custom_element = {custom_element}
}};

// End of code for element {factory_name}
"""



# Copyright (C) 2012-2017 Marie Rognes
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Utility functions for some code shared between representations."""

import hashlib
import logging

import numpy

import ufl
from ffcx.element_interface import (create_quadrature, map_facet_points,
                                    reference_cell_vertices)

logger = logging.getLogger("ffcx")


class QuadratureRule:
    def __init__(self, points, weights):
        self.points = numpy.ascontiguousarray(points)  # TODO: change basix to make this unnecessary
        self.weights = weights
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self.hash_obj = hashlib.sha1(self.points)
            self._hash = int(self.hash_obj.hexdigest(), 32)
        return self._hash

    def __eq__(self, other):
        return numpy.allclose(self.points, other.points) and numpy.allclose(self.weights, other.weights)

    def id(self):
        """Return unique deterministic identifier.

        Note
        ----
        This identifier is used to provide unique names to tables and symbols
        in generated code.

        """
        return self.hash_obj.hexdigest()[-3:]


def create_quadrature_points_and_weights(integral_type, cell, degree, rule):
    """Create quadrature rule and return points and weights."""
    if integral_type == "cell":
        return create_quadrature(cell.cellname(), degree, rule)
    elif integral_type in ufl.measure.facet_integral_types:
        facet_types = cell.facet_types()
        # Raise exception for cells with more than one facet type e.g. prisms
        if len(facet_types) > 1:
            raise Exception(f"Cell type {cell} not supported for integral type {integral_type}.")
        return create_quadrature(facet_types[0].cellname(), degree, rule)
    elif integral_type in ufl.measure.point_integral_types:
        return create_quadrature("vertex", degree, rule)
    elif integral_type == "expression":
        return (None, None)

    logging.exception(f"Unknown integral type: {integral_type}")
    return (None, None)


def integral_type_to_entity_dim(integral_type, tdim):
    """Given integral_type and domain tdim, return the tdim of the integration entity."""
    if integral_type == "cell":
        entity_dim = tdim
    elif integral_type in ufl.measure.facet_integral_types:
        entity_dim = tdim - 1
    elif integral_type in ufl.measure.point_integral_types:
        entity_dim = 0
    elif integral_type in ufl.custom_integral_types:
        entity_dim = tdim
    elif integral_type == "expression":
        entity_dim = tdim
    else:
        raise RuntimeError(f"Unknown integral_type: {integral_type}")
    return entity_dim


def map_integral_points(points, integral_type, cell, entity):
    """Map points from reference entity to its parent reference cell."""
    tdim = cell.topological_dimension()
    entity_dim = integral_type_to_entity_dim(integral_type, tdim)
    if entity_dim == tdim:
        assert points.shape[1] == tdim
        assert entity == 0
        return numpy.asarray(points)
    elif entity_dim == tdim - 1:
        assert points.shape[1] == tdim - 1
        return numpy.asarray(map_facet_points(points, entity, cell.cellname()))
    elif entity_dim == 0:
        return numpy.asarray([reference_cell_vertices(cell.cellname())[entity]])
    else:
        raise RuntimeError(f"Can't map points from entity_dim={entity_dim}")



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Collection of FFCx specific pieces for the code generation phase."""

import types

import ffcx.codegeneration.C.cnodes
from ffcx.codegeneration.access import FFCXBackendAccess
from ffcx.codegeneration.C.ufl_to_cnodes import UFL2CNodesTranslatorCpp
from ffcx.codegeneration.definitions import FFCXBackendDefinitions
from ffcx.codegeneration.symbols import FFCXBackendSymbols


class FFCXBackend(object):
    """Class collecting all aspects of the FFCx backend."""

    def __init__(self, ir, options):

        # This is the seam where cnodes/C is chosen for the FFCx backend
        self.language: types.ModuleType = ffcx.codegeneration.C.cnodes
        scalar_type = options["scalar_type"]
        self.ufl_to_language = UFL2CNodesTranslatorCpp(self.language, scalar_type)

        coefficient_numbering = ir.coefficient_numbering
        coefficient_offsets = ir.coefficient_offsets

        original_constant_offsets = ir.original_constant_offsets

        self.symbols = FFCXBackendSymbols(self.language, coefficient_numbering,
                                          coefficient_offsets, original_constant_offsets)
        self.definitions = FFCXBackendDefinitions(ir, self.language,
                                                  self.symbols, options)
        self.access = FFCXBackendAccess(ir, self.language, self.symbols,
                                        options)



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""FFCx/UFC specific variable access."""

import logging
import warnings

import ufl
from basix.ufl_wrapper import BlockedElement
from ffcx.element_interface import convert_element, create_element

logger = logging.getLogger("ffcx")


class FFCXBackendAccess(object):
    """FFCx specific cpp formatter class."""

    def __init__(self, ir, language, symbols, options):

        # Store ir and options
        self.entitytype = ir.entitytype
        self.integral_type = ir.integral_type
        self.language = language
        self.symbols = symbols
        self.options = options

        # Lookup table for handler to call when the "get" method (below) is
        # called, depending on the first argument type.
        self.call_lookup = {ufl.coefficient.Coefficient: self.coefficient,
                            ufl.constant.Constant: self.constant,
                            ufl.geometry.Jacobian: self.jacobian,
                            ufl.geometry.CellCoordinate: self.cell_coordinate,
                            ufl.geometry.FacetCoordinate: self.facet_coordinate,
                            ufl.geometry.CellVertices: self.cell_vertices,
                            ufl.geometry.FacetEdgeVectors: self.facet_edge_vectors,
                            ufl.geometry.CellEdgeVectors: self.cell_edge_vectors,
                            ufl.geometry.CellFacetJacobian: self.cell_facet_jacobian,
                            ufl.geometry.ReferenceCellVolume: self.reference_cell_volume,
                            ufl.geometry.ReferenceFacetVolume: self.reference_facet_volume,
                            ufl.geometry.ReferenceCellEdgeVectors: self.reference_cell_edge_vectors,
                            ufl.geometry.ReferenceFacetEdgeVectors: self.reference_facet_edge_vectors,
                            ufl.geometry.ReferenceNormal: self.reference_normal,
                            ufl.geometry.CellOrientation: self._pass,
                            ufl.geometry.FacetOrientation: self.facet_orientation,
                            ufl.geometry.SpatialCoordinate: self.spatial_coordinate}

    def get(self, e, mt, tabledata, num_points):
        # Call appropriate handler, depending on the type of e
        handler = self.call_lookup.get(type(e), False)

        if not handler:
            # Look for parent class types instead
            for k in self.call_lookup.keys():
                if isinstance(e, k):
                    handler = self.call_lookup[k]
                    break

        if handler:
            return handler(e, mt, tabledata, num_points)
        else:
            raise RuntimeError(f"Not handled: {type(e)}")

    def coefficient(self, e, mt, tabledata, num_points):
        ttype = tabledata.ttype
        assert ttype != "zeros"

        num_dofs = tabledata.values.shape[3]
        begin = tabledata.offset
        end = begin + tabledata.block_size * (num_dofs - 1) + 1

        if ttype == "ones" and (end - begin) == 1:
            # f = 1.0 * f_{begin}, just return direct reference to dof
            # array at dof begin (if mt is restricted, begin contains
            # cell offset)
            return self.symbols.coefficient_dof_access(mt.terminal, begin)
        else:
            # Return symbol, see definitions for computation
            return self.symbols.coefficient_value(mt)

    def constant(self, e, mt, tabledata, num_points):
        """Access to a constant is handled trivially, directly through constants symbol."""
        return self.symbols.constant_index_access(mt.terminal, mt.flat_component)

    def spatial_coordinate(self, e, mt, tabledata, num_points):
        if mt.global_derivatives:
            raise RuntimeError("Not expecting global derivatives of SpatialCoordinate.")
        if mt.averaged is not None:
            raise RuntimeError("Not expecting average of SpatialCoordinates.")

        if self.integral_type in ufl.custom_integral_types:
            if mt.local_derivatives:
                raise RuntimeError("FIXME: Jacobian in custom integrals is not implemented.")

            # Access predefined quadrature points table
            x = self.symbols.custom_points_table()
            iq = self.symbols.quadrature_loop_index()
            gdim, = mt.terminal.ufl_shape
            if gdim == 1:
                index = iq
            else:
                index = iq * gdim + mt.flat_component
            return x[index]
        elif self.integral_type == "expression":
            # Physical coordinates are computed by code generated in
            # definitions
            return self.symbols.x_component(mt)
        else:
            # Physical coordinates are computed by code generated in
            # definitions
            return self.symbols.x_component(mt)

    def cell_coordinate(self, e, mt, tabledata, num_points):
        if mt.global_derivatives:
            raise RuntimeError("Not expecting derivatives of CellCoordinate.")
        if mt.local_derivatives:
            raise RuntimeError("Not expecting derivatives of CellCoordinate.")
        if mt.averaged is not None:
            raise RuntimeError("Not expecting average of CellCoordinate.")

        if self.integral_type == "cell" and not mt.restriction:
            # Access predefined quadrature points table
            X = self.symbols.points_table(num_points)
            tdim, = mt.terminal.ufl_shape
            iq = self.symbols.quadrature_loop_index()
            if num_points == 1:
                index = mt.flat_component
            elif tdim == 1:
                index = iq
            else:
                index = iq * tdim + mt.flat_component
            return X[index]
        else:
            # X should be computed from x or Xf symbolically instead of
            # getting here
            raise RuntimeError("Expecting reference cell coordinate to be symbolically rewritten.")

    def facet_coordinate(self, e, mt, tabledata, num_points):
        L = self.language
        if mt.global_derivatives:
            raise RuntimeError("Not expecting derivatives of FacetCoordinate.")
        if mt.local_derivatives:
            raise RuntimeError("Not expecting derivatives of FacetCoordinate.")
        if mt.averaged is not None:
            raise RuntimeError("Not expecting average of FacetCoordinate.")
        if mt.restriction:
            raise RuntimeError("Not expecting restriction of FacetCoordinate.")

        if self.integral_type in ("interior_facet", "exterior_facet"):
            tdim, = mt.terminal.ufl_shape
            if tdim == 0:
                raise RuntimeError("Vertices have no facet coordinates.")
            elif tdim == 1:
                warnings.warn(
                    "Vertex coordinate is always 0, should get rid of this in ufl geometry lowering."
                )
                return L.LiteralFloat(0.0)
            Xf = self.points_table(num_points)
            iq = self.symbols.quadrature_loop_index()
            assert 0 <= mt.flat_component < (tdim - 1)
            if num_points == 1:
                index = mt.flat_component
            elif tdim == 2:
                index = iq
            else:
                index = iq * (tdim - 1) + mt.flat_component
            return Xf[index]
        else:
            # Xf should be computed from X or x symbolically instead of
            # getting here
            raise RuntimeError("Expecting reference facet coordinate to be symbolically rewritten.")

    def jacobian(self, e, mt, tabledata, num_points):
        if mt.averaged is not None:
            raise RuntimeError("Not expecting average of Jacobian.")
        return self.symbols.J_component(mt)

    def reference_cell_volume(self, e, mt, tabledata, access):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            return L.Symbol(f"{cellname}_reference_cell_volume")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def reference_facet_volume(self, e, mt, tabledata, access):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            return L.Symbol(f"{cellname}_reference_facet_volume")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def reference_normal(self, e, mt, tabledata, access):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("interval", "triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            table = L.Symbol(f"{cellname}_reference_facet_normals")
            facet = self.symbols.entity("facet", mt.restriction)
            return table[facet][mt.component[0]]
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def cell_facet_jacobian(self, e, mt, tabledata, num_points):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            table = L.Symbol(f"{cellname}_reference_facet_jacobian")
            facet = self.symbols.entity("facet", mt.restriction)
            return table[facet][mt.component[0]][mt.component[1]]
        elif cellname == "interval":
            raise RuntimeError("The reference facet jacobian doesn't make sense for interval cell.")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def reference_cell_edge_vectors(self, e, mt, tabledata, num_points):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            table = L.Symbol(f"{cellname}_reference_edge_vectors")
            return table[mt.component[0]][mt.component[1]]
        elif cellname == "interval":
            raise RuntimeError("The reference cell edge vectors doesn't make sense for interval cell.")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def reference_facet_edge_vectors(self, e, mt, tabledata, num_points):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname in ("tetrahedron", "hexahedron"):
            table = L.Symbol(f"{cellname}_reference_edge_vectors")
            facet = self.symbols.entity("facet", mt.restriction)
            return table[facet][mt.component[0]][mt.component[1]]
        elif cellname in ("interval", "triangle", "quadrilateral"):
            raise RuntimeError(
                "The reference cell facet edge vectors doesn't make sense for interval or triangle cell."
            )
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

    def facet_orientation(self, e, mt, tabledata, num_points):
        L = self.language
        cellname = ufl.domain.extract_unique_domain(mt.terminal).ufl_cell().cellname()
        if cellname not in ("interval", "triangle", "tetrahedron"):
            raise RuntimeError(f"Unhandled cell types {cellname}.")

        table = L.Symbol(f"{cellname}_facet_orientations")
        facet = self.symbols.entity("facet", mt.restriction)
        return table[facet]

    def cell_vertices(self, e, mt, tabledata, num_points):
        # Get properties of domain
        domain = ufl.domain.extract_unique_domain(mt.terminal)
        gdim = domain.geometric_dimension()
        coordinate_element = convert_element(domain.ufl_coordinate_element())

        # Get dimension and dofmap of scalar element
        assert isinstance(coordinate_element, BlockedElement)
        assert coordinate_element.value_shape() == (gdim, )
        ufl_scalar_element, = set(coordinate_element.sub_elements())
        scalar_element = create_element(ufl_scalar_element)
        assert scalar_element.value_size == 1 and scalar_element.block_size == 1

        vertex_scalar_dofs = scalar_element.entity_dofs[0]
        num_scalar_dofs = scalar_element.dim

        # Get dof and component
        dof, = vertex_scalar_dofs[mt.component[0]]
        component = mt.component[1]

        expr = self.symbols.domain_dof_access(dof, component, gdim, num_scalar_dofs, mt.restriction)
        return expr

    def cell_edge_vectors(self, e, mt, tabledata, num_points):
        # Get properties of domain
        domain = ufl.domain.extract_unique_domain(mt.terminal)
        cellname = domain.ufl_cell().cellname()
        gdim = domain.geometric_dimension()
        coordinate_element = convert_element(domain.ufl_coordinate_element())

        if cellname in ("triangle", "tetrahedron", "quadrilateral", "hexahedron"):
            pass
        elif cellname == "interval":
            raise RuntimeError("The physical cell edge vectors doesn't make sense for interval cell.")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

        # Get dimension and dofmap of scalar element
        assert isinstance(coordinate_element, BlockedElement)
        assert coordinate_element.value_shape() == (gdim, )
        ufl_scalar_element, = set(coordinate_element.sub_elements())
        scalar_element = create_element(ufl_scalar_element)
        assert scalar_element.value_size == 1 and scalar_element.block_size == 1

        vertex_scalar_dofs = scalar_element.entity_dofs[0]
        num_scalar_dofs = scalar_element.dim

        # Get edge vertices
        edge = mt.component[0]
        vertex0, vertex1 = scalar_element.reference_topology[1][edge]

        # Get dofs and component
        dof0, = vertex_scalar_dofs[vertex0]
        dof1, = vertex_scalar_dofs[vertex1]
        component = mt.component[1]

        return self.symbols.domain_dof_access(
            dof0, component, gdim, num_scalar_dofs, mt.restriction
        ) - self.symbols.domain_dof_access(
            dof1, component, gdim, num_scalar_dofs, mt.restriction
        )

    def facet_edge_vectors(self, e, mt, tabledata, num_points):
        L = self.language

        # Get properties of domain
        domain = ufl.domain.extract_unique_domain(mt.terminal)
        cellname = domain.ufl_cell().cellname()
        gdim = domain.geometric_dimension()
        coordinate_element = convert_element(domain.ufl_coordinate_element())

        if cellname in ("tetrahedron", "hexahedron"):
            pass
        elif cellname in ("interval", "triangle", "quadrilateral"):
            raise RuntimeError(
                f"The physical facet edge vectors doesn't make sense for {cellname} cell.")
        else:
            raise RuntimeError(f"Unhandled cell types {cellname}.")

        # Get dimension and dofmap of scalar element
        assert isinstance(coordinate_element, BlockedElement)
        assert coordinate_element.value_shape() == (gdim, )
        ufl_scalar_element, = set(coordinate_element.sub_elements())
        scalar_element = create_element(ufl_scalar_element)
        assert scalar_element.value_size == 1 and scalar_element.block_size == 1

        scalar_element = create_element(ufl_scalar_element)
        num_scalar_dofs = scalar_element.dim

        # Get edge vertices
        facet = self.symbols.entity("facet", mt.restriction)
        facet_edge = mt.component[0]
        facet_edge_vertices = L.Symbol(f"{cellname}_facet_edge_vertices")
        vertex0 = facet_edge_vertices[facet][facet_edge][0]
        vertex1 = facet_edge_vertices[facet][facet_edge][1]

        # Get dofs and component
        component = mt.component[1]
        assert coordinate_element.degree() == 1, "Assuming degree 1 element"
        dof0 = vertex0
        dof1 = vertex1
        expr = (
            self.symbols.domain_dof_access(dof0, component, gdim, num_scalar_dofs, mt.restriction)
            - self.symbols.domain_dof_access(dof1, component, gdim, num_scalar_dofs, mt.restriction))

        return expr

    def _pass(self, *args, **kwargs):
        """Return one."""
        return 1



# Copyright (C) 2009-2018 Anders Logg, Martin Sandve Alnæs and Garth N. Wells
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Note: Most of the code in this file is a direct translation from the
# old implementation in FFC

import logging
import typing

import ffcx.codegeneration.dofmap_template as ufcx_dofmap

logger = logging.getLogger("ffcx")


def tabulate_entity_dofs(L, entity_dofs: typing.List[typing.List[typing.List[int]]],
                         num_dofs_per_entity: typing.List[int]):
    # Output argument array
    dofs = L.Symbol("dofs")

    # Input arguments
    d = L.Symbol("d")
    i = L.Symbol("i")

    # TODO: Removed check for (d <= tdim + 1)
    tdim = len(num_dofs_per_entity) - 1

    # Generate cases for each dimension:
    all_cases = []
    for dim in range(tdim + 1):

        # Ignore if no entities for this dimension
        if num_dofs_per_entity[dim] == 0:
            continue

        # Generate cases for each mesh entity
        cases = []
        for entity in range(len(entity_dofs[dim])):
            casebody = []
            for (j, dof) in enumerate(entity_dofs[dim][entity]):
                casebody += [L.Assign(dofs[j], dof)]
            cases.append((entity, L.StatementList(casebody)))

        # Generate inner switch
        # TODO: Removed check for (i <= num_entities-1)
        inner_switch = L.Switch(i, cases, autoscope=False)
        all_cases.append((dim, inner_switch))

    if all_cases:
        return L.Switch(d, all_cases, autoscope=False)
    else:
        return L.NoOp()


def generator(ir, options):
    """Generate UFC code for a dofmap."""
    logger.info("Generating code for dofmap:")
    logger.info(f"--- num element support dofs: {ir.num_element_support_dofs}")
    logger.info(f"--- name: {ir.name}")

    d = {}

    # Attributes
    d["factory_name"] = ir.name
    d["signature"] = f"\"{ir.signature}\""
    d["num_global_support_dofs"] = ir.num_global_support_dofs
    d["num_element_support_dofs"] = ir.num_element_support_dofs
    d["num_sub_dofmaps"] = ir.num_sub_dofmaps

    import ffcx.codegeneration.C.cnodes as L

    num_entity_dofs = ir.num_entity_dofs + [0, 0, 0, 0]
    num_entity_dofs = num_entity_dofs[: 4]
    d["num_entity_dofs"] = f"num_entity_dofs_{ir.name}"
    d["num_entity_dofs_init"] = L.ArrayDecl("int", f"num_entity_dofs_{ir.name}",
                                            values=num_entity_dofs, sizes=4)

    num_entity_closure_dofs = ir.num_entity_closure_dofs + [0, 0, 0, 0]
    num_entity_closure_dofs = num_entity_closure_dofs[:4]
    d["num_entity_closure_dofs"] = f"num_entity_closure_dofs_{ir.name}"
    d["num_entity_closure_dofs_init"] = L.ArrayDecl("int", f"num_entity_closure_dofs_{ir.name}",
                                                    values=num_entity_closure_dofs, sizes=4)

    d["block_size"] = ir.block_size

    # Functions
    d["tabulate_entity_dofs"] = tabulate_entity_dofs(L, ir.entity_dofs, ir.num_entity_dofs)
    d["tabulate_entity_closure_dofs"] = tabulate_entity_dofs(L, ir.entity_closure_dofs, ir.num_entity_closure_dofs)

    if len(ir.sub_dofmaps) > 0:
        d["sub_dofmaps_initialization"] = L.ArrayDecl(
            "ufcx_dofmap*", f"sub_dofmaps_{ir.name}",
            values=[L.AddressOf(L.Symbol(dofmap)) for dofmap in ir.sub_dofmaps], sizes=len(ir.sub_dofmaps))
        d["sub_dofmaps"] = f"sub_dofmaps_{ir.name}"
    else:
        d["sub_dofmaps_initialization"] = ""
        d["sub_dofmaps"] = "NULL"

    # Check that no keys are redundant or have been missed
    from string import Formatter
    fields = [fname for _, fname, _, _ in Formatter().parse(ufcx_dofmap.factory) if fname]
    # Remove square brackets from any field names
    fields = [f.split("[")[0] for f in fields]
    assert set(fields) == set(
        d.keys()), "Mismatch between keys in template and in formatting dict."

    # Format implementation code
    implementation = ufcx_dofmap.factory.format_map(d)

    # Format declaration
    declaration = ufcx_dofmap.declaration.format(factory_name=ir.name)

    return declaration, implementation



# Copyright (C) 2009-2020 Anders Logg and Michal Habera
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import hashlib
import typing

import numpy
import numpy.typing

import ffcx
import ufl

from .element_interface import convert_element


def compute_signature(ufl_objects: typing.List[
    typing.Union[ufl.Form,
                 ufl.FiniteElementBase,
                 typing.Tuple[ufl.core.expr.Expr, numpy.typing.NDArray[numpy.float64]]]], tag: str) -> str:
    """Compute the signature hash.

    Based on the UFL type of the objects and an additional optional
    'tag'.
    """
    object_signature = ""
    for ufl_object in ufl_objects:
        # Get signature from ufl object
        if isinstance(ufl_object, ufl.Form):
            kind = "form"
            object_signature += ufl_object.signature()
        elif isinstance(ufl_object, ufl.FiniteElementBase):
            object_signature += repr(convert_element(ufl_object))
            kind = "element"
        elif isinstance(ufl_object, tuple) and isinstance(ufl_object[0], ufl.core.expr.Expr):
            expr = ufl_object[0]
            points = ufl_object[1]

            # FIXME Move this to UFL, cache the computation
            coeffs = ufl.algorithms.extract_coefficients(expr)
            consts = ufl.algorithms.analysis.extract_constants(expr)
            args = ufl.algorithms.analysis.extract_arguments(expr)

            rn = dict()
            rn.update(dict((c, i) for i, c in enumerate(coeffs)))
            rn.update(dict((c, i) for i, c in enumerate(consts)))
            rn.update(dict((c, i) for i, c in enumerate(args)))

            domains: typing.List[ufl.Mesh] = []
            for coeff in coeffs:
                domains.append(*coeff.ufl_domains())
            for arg in args:
                domains.append(*arg.ufl_function_space().ufl_domains())
            for gc in ufl.algorithms.analysis.extract_type(expr, ufl.classes.GeometricQuantity):
                domains.append(*gc.ufl_domains())
            for const in consts:
                domains.append(const.ufl_domain())
            domains = ufl.algorithms.analysis.unique_tuple(domains)
            rn.update(dict((d, i) for i, d in enumerate(domains)))

            # Hash on UFL signature and points
            signature = ufl.algorithms.signature.compute_expression_signature(expr, rn)
            object_signature += signature
            object_signature += repr(points)

            kind = "expression"
        else:
            raise RuntimeError(f"Unknown ufl object type {ufl_object.__class__.__name__}")

    # Build combined signature
    signatures = [object_signature, str(ffcx.__version__), ffcx.codegeneration.get_signature(), kind, tag]
    string = ";".join(signatures)
    return hashlib.sha1(string.encode('utf-8')).hexdigest()


def integral_name(original_form, integral_type, form_id, subdomain_id, prefix):
    sig = compute_signature([original_form], str((prefix, integral_type, form_id, subdomain_id)))
    return f"integral_{sig}"


def form_name(original_form, form_id, prefix):
    sig = compute_signature([original_form], str((prefix, form_id)))
    return f"form_{sig}"


def finite_element_name(ufl_element, prefix):
    assert isinstance(ufl_element, ufl.FiniteElementBase)
    sig = compute_signature([convert_element(ufl_element)], prefix)
    return f"element_{sig}"


def dofmap_name(ufl_element, prefix):
    assert isinstance(ufl_element, ufl.FiniteElementBase)
    sig = compute_signature([convert_element(ufl_element)], prefix)
    return f"dofmap_{sig}"


def expression_name(expression, prefix):
    assert isinstance(expression[0], ufl.core.expr.Expr)
    sig = compute_signature([expression], prefix)
    return f"expression_{sig}"


def cdtype_to_numpy(cdtype: str):
    """Map a C data type string NumPy datatype string."""
    if cdtype == "double":
        return "float64"
    elif cdtype == "double _Complex":
        return "complex128"
    elif cdtype == "float":
        return "float32"
    elif cdtype == "float _Complex":
        return "complex64"
    elif cdtype == "long double":
        return "longdouble"
    else:
        raise RuntimeError(f"Unknown NumPy type for: {cdtype}")


def scalar_to_value_type(scalar_type: str) -> str:
    """The C value type associated with a C scalar type.

    Args:
      scalar_type: A C type.

    Returns:
      The value type associated with ``scalar_type``. E.g., if
      ``scalar_type`` is ``float _Complex`` the return value is 'float'.

    """
    return scalar_type.replace(' _Complex', '')



# Copyright (C) 2009 Kristian B. Oelgaard
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# Test form for metadata.
from ufl import (Coefficient, FiniteElement, TestFunction, TrialFunction,
                 VectorElement, dx, grad, inner, triangle)

element = FiniteElement("Lagrange", triangle, 1)
vector_element = VectorElement("Lagrange", triangle, 1)


u = TrialFunction(element)
v = TestFunction(element)
c = Coefficient(vector_element)

# Terms on the same subdomain using different quadrature degree
a = inner(grad(u), grad(v)) * dx(0, degree=8)\
    + inner(c, c) * inner(grad(u), grad(v)) * dx(1, degree=4)\
    + inner(c, c) * inner(grad(u), grad(v)) * dx(1, degree=2)\
    + inner(grad(u), grad(v)) * dx(1, degree=-1)



# Copyright (C) 2009-2010 Anders Logg
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2009-03-20
# Last changed: 2011-03-08
#
# Simple example of a form defined over exterior and interior facets.
from ufl import (FacetNormal, FiniteElement, TestFunction, TrialFunction, avg,
                 ds, dS, grad, inner, jump, triangle)

element = FiniteElement("Discontinuous Lagrange", triangle, 1)

u = TrialFunction(element)
v = TestFunction(element)

n = FacetNormal(triangle)

a = u * v * ds \
    + u('+') * v('-') * dS \
    + inner(jump(u, n), avg(grad(v))) * dS \
    + inner(avg(grad(u)), jump(v, n)) * dS



# Copyright (C) 2011-2017 Martin Sandve Alnæs
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""FFCx/UFC specific variable definitions."""

import logging

import ufl
from ffcx.element_interface import create_element
from ffcx.naming import scalar_to_value_type

logger = logging.getLogger("ffcx")


class FFCXBackendDefinitions(object):
    """FFCx specific code definitions."""

    def __init__(self, ir, language, symbols, options):
        # Store ir and options
        self.integral_type = ir.integral_type
        self.entitytype = ir.entitytype
        self.language = language
        self.symbols = symbols
        self.options = options

        self.ir = ir

        # Lookup table for handler to call when the "get" method (below) is
        # called, depending on the first argument type.
        self.call_lookup = {ufl.coefficient.Coefficient: self.coefficient,
                            ufl.constant.Constant: self.constant,
                            ufl.geometry.Jacobian: self.jacobian,
                            ufl.geometry.CellVertices: self._expect_physical_coords,
                            ufl.geometry.FacetEdgeVectors: self._expect_physical_coords,
                            ufl.geometry.CellEdgeVectors: self._expect_physical_coords,
                            ufl.geometry.CellFacetJacobian: self._expect_table,
                            ufl.geometry.ReferenceCellVolume: self._expect_table,
                            ufl.geometry.ReferenceFacetVolume: self._expect_table,
                            ufl.geometry.ReferenceCellEdgeVectors: self._expect_table,
                            ufl.geometry.ReferenceFacetEdgeVectors: self._expect_table,
                            ufl.geometry.ReferenceNormal: self._expect_table,
                            ufl.geometry.CellOrientation: self._pass,
                            ufl.geometry.FacetOrientation: self._expect_table,
                            ufl.geometry.SpatialCoordinate: self.spatial_coordinate}

    def get(self, t, mt, tabledata, quadrature_rule, access):
        # Call appropriate handler, depending on the type of t
        ttype = type(t)
        handler = self.call_lookup.get(ttype, False)

        if not handler:
            # Look for parent class types instead
            for k in self.call_lookup.keys():
                if isinstance(t, k):
                    handler = self.call_lookup[k]
                    break

        if handler:
            return handler(t, mt, tabledata, quadrature_rule, access)
        else:
            raise RuntimeError("Not handled: %s", ttype)

    def coefficient(self, t, mt, tabledata, quadrature_rule, access):
        """Return definition code for coefficients."""
        L = self.language

        ttype = tabledata.ttype
        num_dofs = tabledata.values.shape[3]
        bs = tabledata.block_size
        begin = tabledata.offset
        end = begin + bs * (num_dofs - 1) + 1

        if ttype == "zeros":
            logging.debug("Not expecting zero coefficients to get this far.")
            return [], []

        # For a constant coefficient we reference the dofs directly, so no definition needed
        if ttype == "ones" and end - begin == 1:
            return [], []

        assert begin < end

        # Get access to element table
        FE = self.symbols.element_table(tabledata, self.entitytype, mt.restriction)
        ic = self.symbols.coefficient_dof_sum_index()

        code = []
        pre_code = []

        if bs > 1 and not tabledata.is_piecewise:
            # For bs > 1, the coefficient access has a stride of bs. e.g.: XYZXYZXYZ
            # When memory access patterns are non-sequential, the number of cache misses increases.
            # In turn, it results in noticeably reduced performance.
            # In this case, we create temp arrays outside the quadrature to store the coefficients and
            # have a sequential access pattern.
            dof_access, dof_access_map = self.symbols.coefficient_dof_access_blocked(mt.terminal, ic, bs, begin)

            # If a map is necessary from stride 1 to bs, the code must be added before the quadrature loop.
            if dof_access_map:
                pre_code += [L.ArrayDecl(self.options["scalar_type"], dof_access.array, num_dofs)]
                pre_body = L.Assign(dof_access, dof_access_map)
                pre_code += [L.ForRange(ic, 0, num_dofs, pre_body)]
        else:
            dof_access = self.symbols.coefficient_dof_access(mt.terminal, ic * bs + begin)

        body = [L.AssignAdd(access, dof_access * FE[ic])]
        code += [L.VariableDecl(self.options["scalar_type"], access, 0.0)]
        code += [L.ForRange(ic, 0, num_dofs, body)]

        return pre_code, code

    def constant(self, t, mt, tabledata, quadrature_rule, access):
        # Constants are not defined within the kernel.
        # No definition is needed because access to them is directly
        # via symbol c[], i.e. as passed into the kernel.
        return [], []

    def _define_coordinate_dofs_lincomb(self, e, mt, tabledata, quadrature_rule, access):
        """Define x or J as a linear combination of coordinate dofs with given table data."""
        L = self.language

        # Get properties of domain
        domain = ufl.domain.extract_unique_domain(mt.terminal)
        coordinate_element = domain.ufl_coordinate_element()
        num_scalar_dofs = create_element(coordinate_element).sub_element.dim

        num_dofs = tabledata.values.shape[3]
        begin = tabledata.offset

        assert num_scalar_dofs == num_dofs

        # Find table name
        ttype = tabledata.ttype

        assert ttype != "zeros"
        assert ttype != "ones"

        # Get access to element table
        FE = self.symbols.element_table(tabledata, self.entitytype, mt.restriction)
        ic = self.symbols.coefficient_dof_sum_index()
        dof_access = self.symbols.S("coordinate_dofs")

        # coordinate dofs is always 3d
        dim = 3
        offset = 0
        if mt.restriction == "-":
            offset = num_scalar_dofs * dim

        value_type = scalar_to_value_type(self.options["scalar_type"])

        code = []
        body = [L.AssignAdd(access, dof_access[ic * dim + begin + offset] * FE[ic])]
        code += [L.VariableDecl(f"{value_type}", access, 0.0)]
        code += [L.ForRange(ic, 0, num_scalar_dofs, body)]

        return [], code

    def spatial_coordinate(self, e, mt, tabledata, quadrature_rule, access):
        """Return definition code for the physical spatial coordinates.

        If physical coordinates are given:
          No definition needed.

        If reference coordinates are given:
          x = sum_k xdof_k xphi_k(X)

        If reference facet coordinates are given:
          x = sum_k xdof_k xphi_k(Xf)
        """
        if self.integral_type in ufl.custom_integral_types:
            # FIXME: Jacobian may need adjustment for custom_integral_types
            if mt.local_derivatives:
                logging.exception("FIXME: Jacobian in custom integrals is not implemented.")
            return []
        else:
            return self._define_coordinate_dofs_lincomb(e, mt, tabledata, quadrature_rule, access)

    def jacobian(self, e, mt, tabledata, quadrature_rule, access):
        """Return definition code for the Jacobian of x(X)."""
        return self._define_coordinate_dofs_lincomb(e, mt, tabledata, quadrature_rule, access)

    def _expect_table(self, e, mt, tabledata, quadrature_rule, access):
        """Return quantities referring to constant tables defined in the generated code."""
        # TODO: Inject const static table here instead?
        return [], []

    def _expect_physical_coords(self, e, mt, tabledata, quadrature_rule, access):
        """Return quantities referring to coordinate_dofs."""
        # TODO: Generate more efficient inline code for Max/MinCell/FacetEdgeLength
        #       and CellDiameter here rather than lowering these quantities?
        return [], []

    def _pass(self, *args, **kwargs):
        """Return nothing."""
        return [], []



from ufl import (FiniteElement, MixedElement, TestFunctions, TrialFunctions,
                 ds, grad, inner, triangle)

element1 = FiniteElement("DG", triangle, 1)
element2 = FiniteElement("DGT", triangle, 1)
element = MixedElement(element1, element2)

u = TrialFunctions(element)[0]
v = TestFunctions(element)[0]

a = inner(grad(u), grad(v)) * ds



# Copyright (C) 2020 Matthew Scroggs
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy as np

import ffcx
import ffcx.codegeneration.jit
import ufl


def test_finite_element(compile_args):
    ufl_element = ufl.FiniteElement("Lagrange", ufl.triangle, 1)
    jit_compiled_elements, module, code = ffcx.codegeneration.jit.compile_elements(
        [ufl_element], cffi_extra_compile_args=compile_args)
    ufcx_element, ufcx_dofmap = jit_compiled_elements[0]

    assert ufcx_element.topological_dimension == 2
    assert ufcx_element.geometric_dimension == 2
    assert ufcx_element.space_dimension == 3
    assert ufcx_element.value_rank == 0
    assert ufcx_element.value_size == 1
    assert ufcx_element.reference_value_rank == 0
    assert ufcx_element.reference_value_size == 1
    assert ufcx_element.block_size == 1
    assert ufcx_element.num_sub_elements == 0

    assert ufcx_dofmap.block_size == 1
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_element_support_dofs == 3
    assert ufcx_dofmap.num_entity_dofs[0] == 1
    assert ufcx_dofmap.num_entity_dofs[1] == 0
    assert ufcx_dofmap.num_entity_dofs[2] == 0
    assert ufcx_dofmap.num_entity_dofs[3] == 0
    for v in range(3):
        vals = np.zeros(1, dtype=np.int32)
        vals_ptr = module.ffi.cast("int *", module.ffi.from_buffer(vals))
        ufcx_dofmap.tabulate_entity_dofs(vals_ptr, 0, v)
        assert vals[0] == v
    assert ufcx_dofmap.num_sub_dofmaps == 0


def test_vector_element(compile_args):
    ufl_element = ufl.VectorElement("Lagrange", ufl.triangle, 1)
    jit_compiled_elements, module, code = ffcx.codegeneration.jit.compile_elements(
        [ufl_element], cffi_extra_compile_args=compile_args)
    ufcx_element, ufcx_dofmap = jit_compiled_elements[0]

    assert ufcx_element.topological_dimension == 2
    assert ufcx_element.geometric_dimension == 2
    assert ufcx_element.space_dimension == 6
    assert ufcx_element.value_rank == 1
    assert ufcx_element.value_shape[0] == 2
    assert ufcx_element.value_size == 2
    assert ufcx_element.reference_value_rank == 1
    assert ufcx_element.reference_value_shape[0] == 2
    assert ufcx_element.reference_value_size == 2
    assert ufcx_element.block_size == 2
    assert ufcx_element.num_sub_elements == 2

    assert ufcx_dofmap.block_size == 2
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_element_support_dofs == 3
    assert ufcx_dofmap.num_entity_dofs[0] == 1
    assert ufcx_dofmap.num_entity_dofs[1] == 0
    assert ufcx_dofmap.num_entity_dofs[2] == 0
    assert ufcx_dofmap.num_entity_dofs[3] == 0
    for v in range(3):
        vals = np.zeros(1, dtype=np.int32)
        vals_ptr = module.ffi.cast("int *", module.ffi.from_buffer(vals))
        ufcx_dofmap.tabulate_entity_dofs(vals_ptr, 0, v)
        assert vals[0] == v
    assert ufcx_dofmap.num_sub_dofmaps == 2


def test_tensor_element(compile_args):
    ufl_element = ufl.TensorElement("Lagrange", ufl.triangle, 1)
    jit_compiled_elements, module, code = ffcx.codegeneration.jit.compile_elements(
        [ufl_element], cffi_extra_compile_args=compile_args)
    ufcx_element, ufcx_dofmap = jit_compiled_elements[0]

    assert ufcx_element.topological_dimension == 2
    assert ufcx_element.geometric_dimension == 2
    assert ufcx_element.space_dimension == 12
    assert ufcx_element.value_rank == 2
    assert ufcx_element.value_shape[0] == 2
    assert ufcx_element.value_shape[1] == 2
    assert ufcx_element.value_size == 4
    assert ufcx_element.reference_value_rank == 2
    assert ufcx_element.reference_value_shape[0] == 2
    assert ufcx_element.reference_value_shape[1] == 2
    assert ufcx_element.reference_value_size == 4
    assert ufcx_element.block_size == 4
    assert ufcx_element.num_sub_elements == 4

    assert ufcx_dofmap.block_size == 4
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_global_support_dofs == 0
    assert ufcx_dofmap.num_element_support_dofs == 3
    assert ufcx_dofmap.num_entity_dofs[0] == 1
    assert ufcx_dofmap.num_entity_dofs[1] == 0
    assert ufcx_dofmap.num_entity_dofs[2] == 0
    assert ufcx_dofmap.num_entity_dofs[3] == 0
    for v in range(3):
        vals = np.zeros(1, dtype=np.int32)
        vals_ptr = module.ffi.cast("int *", module.ffi.from_buffer(vals))
        ufcx_dofmap.tabulate_entity_dofs(vals_ptr, 0, v)
        assert vals[0] == v
    assert ufcx_dofmap.num_sub_dofmaps == 4



# Copyright (C) 2018 Chris N. Richardson
#
# This file is part of FFCx. (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import os
import os.path
import subprocess


def test_cmdline_simple():
    os.chdir(os.path.dirname(__file__))
    subprocess.run(["ffcx", "Poisson.py"])


def test_visualise():
    os.chdir(os.path.dirname(__file__))
    subprocess.run(["ffcx", "--visualise", "Poisson.py"])
    assert os.path.isfile("S.pdf")
    assert os.path.isfile("F.pdf")



# Copyright (C) 2011-2017 Martin Sandve Alnæs and Chris Richardson
#
# This file is part of FFCx.(https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import ufl


def handle_scalar_nary(o, ops):
    if o.ufl_shape != ():
        raise RuntimeError("Expecting scalar.")
    sops = [op[0] for op in ops]
    return [o._ufl_expr_reconstruct_(*sops)]


def handle_condition(o, ops):
    # A condition is always scalar, so len(op) == 1
    sops = [op[0] for op in ops]
    return [o._ufl_expr_reconstruct_(*sops)]


def handle_conditional(o, ops):
    # A condition can be non scalar
    symbols = []
    n = len(ops[1])
    if len(ops[0]) != 1:
        raise RuntimeError("Condition should be scalar.")
    if n != len(ops[2]):
        raise RuntimeError("Conditional branches should have same shape.")
    for i in range(len(ops[1])):
        sops = (ops[0][0], ops[1][i], ops[2][i])
        symbols.append(o._ufl_expr_reconstruct_(*sops))
    return symbols


def handle_elementwise_unary(o, ops):
    if len(ops) > 1:
        raise RuntimeError("Expecting unary operator.")
    return [o._ufl_expr_reconstruct_(op) for op in ops[0]]


def handle_division(o, ops):
    if len(ops) != 2:
        raise RuntimeError("Expecting two operands.")
    if len(ops[1]) != 1:
        raise RuntimeError("Expecting scalar divisor.")
    b, = ops[1]
    return [o._ufl_expr_reconstruct_(a, b) for a in ops[0]]


def handle_sum(o, ops):
    if len(ops) != 2:
        raise RuntimeError("Expecting two operands.")
    if len(ops[0]) != len(ops[1]):
        raise RuntimeError("Expecting scalar divisor.")
    return [o._ufl_expr_reconstruct_(a, b) for a, b in zip(ops[0], ops[1])]


def handle_product(o, ops):
    if len(ops) != 2:
        raise RuntimeError("Expecting two operands.")

    # Get the simple cases out of the way
    if len(ops[0]) == 1:  # True scalar * something
        a, = ops[0]
        return [ufl.classes.Product(a, b) for b in ops[1]]
    elif len(ops[1]) == 1:  # Something * true scalar
        b, = ops[1]
        return [ufl.classes.Product(a, b) for a in ops[0]]

    # Neither of operands are true scalars, this is the tricky part
    o0, o1 = o.ufl_operands

    # Get shapes and index shapes
    fi = o.ufl_free_indices
    fi0 = o0.ufl_free_indices
    fi1 = o1.ufl_free_indices
    fid = o.ufl_index_dimensions
    fid0 = o0.ufl_index_dimensions
    fid1 = o1.ufl_index_dimensions

    # Need to map each return component to one component of o0 and
    # one component of o1
    indices = ufl.permutation.compute_indices(fid)

    # Compute which component of o0 is used in component (comp,ind) of o
    # Compute strides within free index spaces
    ist0 = ufl.utils.indexflattening.shape_to_strides(fid0)
    ist1 = ufl.utils.indexflattening.shape_to_strides(fid1)
    # Map o0 and o1 indices to o indices
    indmap0 = [fi.index(i) for i in fi0]
    indmap1 = [fi.index(i) for i in fi1]
    indks = [(ufl.utils.indexflattening.flatten_multiindex([ind[i] for i in indmap0], ist0),
              ufl.utils.indexflattening.flatten_multiindex([ind[i] for i in indmap1], ist1))
             for ind in indices]

    # Build products for scalar components
    results = [ufl.classes.Product(ops[0][k0], ops[1][k1]) for k0, k1 in indks]
    return results


def handle_index_sum(o, ops):
    summand, mi = o.ufl_operands
    ic = mi[0].count()
    fi = summand.ufl_free_indices
    fid = summand.ufl_index_dimensions
    ipos = fi.index(ic)
    d = fid[ipos]

    # Compute "macro-dimensions" before and after i in the total shape of a
    predim = ufl.product(summand.ufl_shape) * ufl.product(fid[:ipos])
    postdim = ufl.product(fid[ipos + 1:])

    # Map each flattened total component of summand to
    # flattened total component of indexsum o by removing
    # axis corresponding to summation index ii.
    ss = ops[0]  # Scalar subexpressions of summand
    if len(ss) != predim * postdim * d:
        raise RuntimeError("Mismatching number of subexpressions.")
    sops = []
    for i in range(predim):
        iind = i * (postdim * d)
        for k in range(postdim):
            ind = iind + k
            sops.append([ss[ind + j * postdim] for j in range(d)])

    # For each scalar output component, sum over collected subcomponents
    # TODO: Need to split this into binary additions to work with future CRSArray format,
    #       i.e. emitting more expressions than there are symbols for this node.
    results = [sum(sop) for sop in sops]
    return results

# TODO: To implement compound tensor operators such as dot and inner,
# we need to identify which index to do the contractions over,
# and build expressions such as sum(a*b for a,b in zip(aops, bops))


_reconstruct_call_lookup = {ufl.classes.MathFunction: handle_scalar_nary,
                            ufl.classes.Abs: handle_scalar_nary,
                            ufl.classes.MinValue: handle_scalar_nary,
                            ufl.classes.MaxValue: handle_scalar_nary,
                            ufl.classes.Real: handle_elementwise_unary,
                            ufl.classes.Imag: handle_elementwise_unary,
                            ufl.classes.Power: handle_scalar_nary,
                            ufl.classes.BesselFunction: handle_scalar_nary,
                            ufl.classes.Atan2: handle_scalar_nary,
                            ufl.classes.Product: handle_product,
                            ufl.classes.Division: handle_division,
                            ufl.classes.Sum: handle_sum,
                            ufl.classes.IndexSum: handle_index_sum,
                            ufl.classes.Conj: handle_elementwise_unary,
                            ufl.classes.Conditional: handle_conditional,
                            ufl.classes.Condition: handle_condition}


def reconstruct(o, *args):
    # First look for exact match
    f = _reconstruct_call_lookup.get(type(o), False)
    if f:
        return f(o, *args)
    else:
        # Look for parent class types instead
        for k in _reconstruct_call_lookup.keys():
            if isinstance(o, k):
                return _reconstruct_call_lookup[k](o, *args)
        # Nothing found
        raise RuntimeError("Not expecting expression of type %s in here." % type(o))



# Copyright (C) 2010 Garth N. Wells
#
# This file is part of FFCx.
#
# FFCx is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFCx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FFCx. If not, see <http://www.gnu.org/licenses/>.
from ufl import (Coefficient, FiniteElement, TestFunction, TrialFunction, avg,
                 derivative, dot, dS, dx, grad, inner, triangle)

element = FiniteElement("Discontinuous Lagrange", triangle, 1)

v = TestFunction(element)
w = Coefficient(element)
L = inner(grad(w), grad(v)) * dx - dot(avg(grad(w)), avg(grad(v))) * dS

u = TrialFunction(element)
a = derivative(L, w, u)



# Code generation format strings for UFC (Unified Form-assembly Code)
# This code is released into the public domain.
#
# The FEniCS Project (http://www.fenicsproject.org/) 2020.

declaration = """
extern ufcx_form {factory_name};

// Helper used to create form using name which was given to the
// form in the UFL file.
// This helper is called in user c++ code.
//
extern ufcx_form* {name_from_uflfile};

// Helper used to create function space using function name
// i.e. name of the Python variable.
//
ufcx_function_space* functionspace_{name_from_uflfile}(const char* function_name);
"""

factory = """
// Code for form {factory_name}

{original_coefficient_position_init}
{dofmaps_init}
{finite_elements_init}

// Return a list of the coefficient names.
const char** coefficient_name_{factory_name}(void)
{{
{coefficient_name_map}
}}

// Return a list of the constant names.
const char** constant_name_{factory_name}(void)
{{
{constant_name_map}
}}

int* integral_ids_{factory_name}(ufcx_integral_type integral_type)
{{
{integral_ids}
}}

int num_integrals_{factory_name}(ufcx_integral_type integral_type)
{{
{num_integrals}
}}

ufcx_integral** integrals_{factory_name}(ufcx_integral_type integral_type)
{{
{integrals}
}}

ufcx_form {factory_name} =
{{

  .signature = {signature},
  .rank = {rank},
  .num_coefficients = {num_coefficients},
  .num_constants = {num_constants},
  .original_coefficient_position = {original_coefficient_position},

  .coefficient_name_map = coefficient_name_{factory_name},
  .constant_name_map = constant_name_{factory_name},

  .finite_elements = {finite_elements},
  .dofmaps = {dofmaps},

  .integral_ids = integral_ids_{factory_name},
  .num_integrals = num_integrals_{factory_name},

  .integrals = integrals_{factory_name}
}};

// Alias name
ufcx_form* {name_from_uflfile} = &{factory_name};

ufcx_function_space* functionspace_{name_from_uflfile}(const char* function_name)
{{
{functionspace}
}}

// End of code for form {factory_name}
"""
